#!/usr/bin/python3

import json
import os
import sys

import gi
gi.require_version('Gdk', '3.0')
gi.require_version('Gspell', '1')
gi.require_version('Gtk', '3.0')
gi.require_version('XApp', '1.0')
from gi.repository import Gdk, Gio, GObject, Gspell, Gtk, Pango, XApp

from xapp.GSettingsWidgets import *

from note_buffer import NoteBuffer
from manager import NotesManager
from common import FileHandler, HoverBox, prompt, confirm
from util import gnote_to_internal_format

import gettext
gettext.install("sticky", "/usr/share/locale", names="ngettext")

APPLICATION_ID = 'org.x.sticky'
STYLE_SHEET_PATH = '/usr/share/sticky/sticky.css'
SCHEMA = 'org.x.sticky'
DBUS_PATH = '/org/x/sticky'
DBUS_INTERFACE_XML = '''
<node>
  <interface name='org.x.sticky'>
    <method name='ShowNotes'>
    </method>
    <method name='NewNote'>
      <arg type='s' name='text' direction='in'/>
    </method>
    <method name='NewNoteBlank'>
    </method>
    <method name='ReloadNotesFromFile'>
    </method>
    <signal name='NotesChanged'>
    </signal>
  </interface>
</node>
'''

UPDATE_DELAY = 1

FONT_SCALES = [
    ('small', _("Small Text"), 'small'),
    ('normal', _("Normal Text"), 'medium'),
    ('large', _("Large Text"), 'large'),
    ('larger', _("Larger Text"), 'x-large')
]

COLORS = {
    'red': _("Red"),
    'green': _("Green"),
    'blue': _("Blue"),
    'yellow': _("Yellow"),
    'purple': _("Purple"),
    'teal': _("Teal"),
    'orange': _("Orange"),
    'magenta': _("Magenta")
}

COLOR_CODES = {
    'red': "#ff5561",
    'green': "#67ff67",
    'blue': "#3d9bff",
    'yellow': "#f6f907",
    'purple': "#a553ff",
    'teal': "#41ffed",
    'orange': "#ffa939",
    'magenta': "#ff7ff7"
}

SHORTCUTS = {
    _("Operations"): [
        (_("Move selection up"), '<ctrl><shift>Up'),
        (_("Move selection down"), '<ctrl><shift>Down'),
        (_("Undo"), '<ctrl>z'),
        (_("Redo"), '<ctrl>y'),
        (_("Toggle Checklist"), '<ctrl>e'),
        (_("Toggle Bullets"), '<ctrl>l')
    ],
    _("Formatting"): [
        (_("Bold"), '<ctrl>b'),
        (_("Italic"), '<ctrl>i'),
        (_("Fixed Width"), '<ctrl>f'),
        (_("Underline"), '<ctrl>u'),
        (_("Strikethrough"), '<ctrl>k'),
        (_("Highlight"), '<ctrl>g'),
        (_("Header"), '<ctrl>h')
    ]
}

START_POSITIONS = {
    'top-left': _("Top Left"),
    'top-center': _("Top Center"),
    'top-right': _("Top Right"),
    'center-left': _("Center Left"),
    'center-center': _("Center"),
    'center-right': _("Center Right"),
    'bottom-left': _("Bottom Left"),
    'bottom-center': _("Bottom Center"),
    'bottom-right': _("Bottom Right"),
}

class Note(Gtk.Window):
    @GObject.Signal(flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    accumulator=GObject.signal_accumulator_true_handled)
    def update(self):
        pass

    @GObject.Signal(flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    accumulator=GObject.signal_accumulator_true_handled)
    def removed(self):
        pass

    def __init__(self, app, parent, info={}):
        self.app = app

        self.showing = False
        self.is_pinned = False
        self.changed_timer_id = 0
        self.invalid_cache = False

        self.x = info.get('x', 0)
        self.y = info.get('y', 0)
        self.height = info.get('height', self.app.settings.get_uint('default-height'))
        self.width = info.get('width', self.app.settings.get_uint('default-width'))
        title = info.get('title', '')
        self.cached_text = info.get('text', '')
        self.color = info.get('color', self.app.settings.get_string('default-color'))

        super(Note, self).__init__(
            skip_taskbar_hint=True,
            transient_for=parent,
            type_hint=Gdk.WindowTypeHint.UTILITY,
            default_height=self.height,
            default_width=self.width,
            resizable=True,
            deletable=False,
            name='sticky-note'
        )

        if self.color == 'cycle':
            if self.app.settings.get_string('last-color') == '':
                last_color = self.color
            else:
                last_color = self.app.settings.get_string('last-color')

            color_keys = list(COLORS)
            try:
                self.color = color_keys[color_keys.index(last_color) + 1]
            except (ValueError, IndexError):
                self.color = color_keys[0]

            self.app.settings.set_string('last-color', self.color)

        context = self.get_style_context()
        context.add_class(self.color)

        if self.app.settings.get_boolean('desktop-window-state'):
            self.stick()

        # title bar
        self.title_bar = Gtk.Box(height_request=30, name='title-bar')
        self.title_bar.connect('button-press-event', self.on_title_click)

        color_icon = Gtk.Image.new_from_icon_name('sticky-color', Gtk.IconSize.BUTTON)
        color_button = Gtk.MenuButton(image=color_icon, relief=Gtk.ReliefStyle.NONE, name='window-button', valign=Gtk.Align.CENTER)
        color_button.connect('button-press-event', self.on_title_click)
        color_button.set_tooltip_text(_("Note Color"))
        self.title_bar.pack_start(color_button, False, False, 0)

        # used to show the edit title icon when the title is hovered
        self.title_hover = HoverBox()
        self.title_bar.pack_start(self.title_hover, True, True, 4)

        self.title_box = Gtk.Box()
        self.title_hover.add(self.title_box)
        self.title = Gtk.Label(label=title, margin_top=4, name='title')
        self.title_box.pack_start(self.title, False, False, 0)
        self.title_style_manager = XApp.StyleManager(widget=self.title_box)

        edit_title_icon = Gtk.Image.new_from_icon_name('sticky-edit', Gtk.IconSize.BUTTON)
        self.edit_title_button = Gtk.Button(image=edit_title_icon, relief=Gtk.ReliefStyle.NONE, name='window-button', valign=Gtk.Align.CENTER)
        self.edit_title_button.connect('clicked', self.set_title)
        self.edit_title_button.connect('button-press-event', self.on_title_click)
        self.edit_title_button.set_tooltip_text(_("Rename"))
        self.title_box.pack_start(self.edit_title_button, False, False, 0)
        self.title_hover.set_child_widget(self.edit_title_button)

        close_icon = Gtk.Image.new_from_icon_name('sticky-delete', Gtk.IconSize.BUTTON)
        close_button = Gtk.Button(image=close_icon, relief=Gtk.ReliefStyle.NONE, name='window-button', valign=Gtk.Align.CENTER)
        close_button.connect('clicked', self.remove)
        close_button.connect('button-press-event', self.on_title_click)
        close_button.set_tooltip_text(_("Delete Note"))
        self.title_bar.pack_end(close_button, False, False, 0)

        add_icon = Gtk.Image.new_from_icon_name('sticky-add', Gtk.IconSize.BUTTON)
        add_button = Gtk.Button(image=add_icon, relief=Gtk.ReliefStyle.NONE, name='window-button', valign=Gtk.Align.CENTER)
        add_button.connect('clicked', self.app.new_note, self)
        add_button.connect('button-press-event', self.on_title_click)
        add_button.set_tooltip_text(_("New Note"))
        self.title_bar.pack_end(add_button, False, False, 0)

        text_icon = Gtk.Image.new_from_icon_name('sticky-text', Gtk.IconSize.BUTTON)
        text_button = Gtk.MenuButton(image=text_icon, relief=Gtk.ReliefStyle.NONE, name='window-button', valign=Gtk.Align.CENTER)
        text_button.connect('button-press-event', self.on_title_click)
        text_button.set_tooltip_text(_("Format"))
        self.title_bar.pack_end(text_button, False, False, 20)

        self.set_titlebar(self.title_bar)

        # buffer
        self.buffer = NoteBuffer()

        # text view
        self.view = Gtk.TextView(wrap_mode=Gtk.WrapMode.WORD_CHAR, populate_all=True, buffer=self.buffer)
        self.buffer.set_view(self.view)
        spell_checker = Gspell.TextView.get_from_gtk_text_view(self.view)
        spell_checker.basic_setup()
        self.app.settings.bind('inline-spell-check', spell_checker, 'inline-spell-checking', Gio.SettingsBindFlags.GET)
        self.view.set_left_margin(10)
        self.view.set_right_margin(10)
        self.view.set_top_margin(10)
        self.view.set_bottom_margin(10)
        self.view.connect('populate-popup', lambda w, p: self.add_context_menu_items(p))
        self.view.connect('key-press-event', self.on_key_press)
        self.view_style_manager = XApp.StyleManager(widget=self.view)

        scroll = Gtk.ScrolledWindow()
        self.add(scroll)
        scroll.add(self.view)

        self.buffer.set_from_internal_markup(self.cached_text)
        self.changed_id = self.buffer.connect('content-changed', self.queue_update, True)

        self.app.settings.connect('changed::font', self.set_font)
        self.set_font()

        self.create_format_menu(color_button, text_button)

        self.connect('configure-event', self.on_size_position_changed)
        self.connect('show', self.on_show)
        self.connect('window-state-event', self.update_window_state)

        self.move(self.x, self.y)

        self.show_all()

    def test(self, *args):
        self.buffer.test()

    def on_size_position_changed(self, *args):
        if self.showing:
            self.showing = False
            return

        (new_x, new_y) = self.get_position()
        (new_width, new_height) = self.get_size()
        if self.x == new_x and self.y == new_y and self.height == new_height and self.width == new_width:
            return

        self.x = new_x
        self.y = new_y
        self.height = new_height
        self.width = new_width

        self.queue_update()

    def on_show(self, *args):
        self.showing = True

    def on_key_press(self, v, event):
        if event.get_state() & Gdk.ModifierType.CONTROL_MASK:
            if event.get_state() & Gdk.ModifierType.SHIFT_MASK:
                if event.get_keyval()[1] == Gdk.KEY_Up:
                    self.buffer.shift(True)
                    return Gdk.EVENT_STOP

                elif event.get_keyval()[1] == Gdk.KEY_Down:
                    self.buffer.shift(False)
                    return Gdk.EVENT_STOP

            if event.get_keyval()[1] == Gdk.KEY_z:
                self.buffer.undo()
                return Gdk.EVENT_STOP

            elif event.get_keyval()[1] == Gdk.KEY_y:
                self.buffer.redo()
                return Gdk.EVENT_STOP

            elif event.get_keyval()[1] == Gdk.KEY_e:
                self.buffer.toggle_checklist()
                return Gdk.EVENT_STOP

            elif event.get_keyval()[1] == Gdk.KEY_l:
                self.buffer.toggle_bullets()
                return Gdk.EVENT_STOP

            elif event.get_keyval()[1] == Gdk.KEY_b:
                self.buffer.tag_selection('bold')
                return Gdk.EVENT_STOP

            elif event.get_keyval()[1] == Gdk.KEY_i:
                self.buffer.tag_selection('italic')
                return Gdk.EVENT_STOP

            elif event.get_keyval()[1] == Gdk.KEY_f:
                self.buffer.tag_selection('monospace')
                return Gdk.EVENT_STOP

            elif event.get_keyval()[1] == Gdk.KEY_u:
                self.buffer.tag_selection('underline')
                return Gdk.EVENT_STOP

            elif event.get_keyval()[1] == Gdk.KEY_h:
                self.buffer.tag_selection('header')
                return Gdk.EVENT_STOP

            elif event.get_keyval()[1] == Gdk.KEY_k:
                self.buffer.tag_selection('strikethrough')
                return Gdk.EVENT_STOP

            elif event.get_keyval()[1] == Gdk.KEY_g:
                self.buffer.tag_selection('highlight')
                return Gdk.EVENT_STOP

            elif event.get_keyval()[1] == Gdk.KEY_2:
                self.buffer.tag_selection('small')
                return Gdk.EVENT_STOP

            elif event.get_keyval()[1] == Gdk.KEY_3:
                self.buffer.tag_selection('normal')
                return Gdk.EVENT_STOP

            elif event.get_keyval()[1] == Gdk.KEY_4:
                self.buffer.tag_selection('large')
                return Gdk.EVENT_STOP

            elif event.get_keyval()[1] == Gdk.KEY_5:
                self.buffer.tag_selection('larger')
                return Gdk.EVENT_STOP

        elif event.keyval in (Gdk.KEY_Return, Gdk.KEY_ISO_Enter, Gdk.KEY_KP_Enter):
            return self.buffer.on_return()

        return Gdk.EVENT_PROPAGATE

    def update_window_state(self, w, event):
        self.is_stuck = event.new_window_state & Gdk.WindowState.STICKY
        # for some reason, the ABOVE flag is never actually being set, even when it should be
        # self.is_pinned = event.new_window_state & Gdk.WindowState.ABOVE

    def on_title_click(self, w, event):
        if event.button == 3:
            menu = Gtk.Menu()
            self.add_context_menu_items(menu, True)
            menu.popup(None, None, None, None, event.button, event.time)

            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def restore(self, time=0):
        if time == 0:
            time = Gtk.get_current_event_time()

        self.show()
        self.present_with_time(time)
        self.move(self.x, self.y)

    def queue_update(self, b=None, invalidate_cache=False):
        self.invalid_cache = invalidate_cache

        if self.changed_timer_id:
            GLib.source_remove(self.changed_timer_id)

        self.changed_timer_id = GLib.timeout_add_seconds(UPDATE_DELAY, self.trigger_update)

    def trigger_update(self):
        self.changed_timer_id = 0

        self.emit('update')

    def get_info(self):
        if self.invalid_cache:
            self.cached_text = self.buffer.get_internal_markup()
            self.invalid_cache = False

        (width, height) = self.get_size()
        info = {
            'x': self.x,
            'y': self.y,
            'height': self.height,
            'width': self.width,
            'color': self.color,
            'title': self.title.get_text(),
            'text': self.cached_text
        }

        return info

    def add_context_menu_items(self, popup, is_title=False):
        if not is_title:
            popup.append(Gtk.SeparatorMenuItem(visible=True))

            self.undo_item = Gtk.MenuItem(label=_("undo"), visible=True, sensitive=self.buffer.can_undo)
            self.undo_item.connect('activate', self.buffer.undo)
            popup.append(self.undo_item)

            self.redo_item = Gtk.MenuItem(label=_("redo"), visible=True, sensitive=self.buffer.can_redo)
            self.redo_item.connect('activate', self.buffer.redo)
            popup.append(self.redo_item)

            popup.append(Gtk.SeparatorMenuItem(visible=True))

        label = _("Set Title") if self.title.get_text() == '' else _('Edit Title')
        edit_title = Gtk.MenuItem(label=label, visible=True)
        edit_title.connect('activate', self.set_title)
        popup.append(edit_title)

        duplicate_item = Gtk.MenuItem(label=_("Duplicate Note"), visible=True)
        duplicate_item.connect('activate', self.duplicate)
        popup.append(duplicate_item)

        remove_item = Gtk.MenuItem(label=_("Delete Note"), visible=True)
        remove_item.connect('activate', self.remove)
        popup.append(remove_item)

        if is_title:
            popup.append(Gtk.SeparatorMenuItem(visible=True))

            if self.is_stuck:
                label = _("Only on This Workspace")
                def on_activate(*args):
                    self.unstick()
            else:
                label = _("Always on Visible Workspace")
                def on_activate(*args):
                    self.stick()

            stick_menu_item = Gtk.MenuItem(label=label, visible=True)
            stick_menu_item.connect('activate', on_activate)
            popup.append(stick_menu_item)

            def on_activate(*args):
                self.set_keep_above(not self.is_pinned)
                self.is_pinned = not self.is_pinned

            pin_menu_item = Gtk.CheckMenuItem(active=self.is_pinned, label=_("Always on Top"), visible=True)
            pin_menu_item.connect('activate', on_activate)
            popup.append(pin_menu_item)

    def create_format_menu(self, color_button, text_button):

        menu = Gtk.Menu()

        for color, color_name in sorted(COLORS.items(), key=lambda item: item[1]):
            color_code = COLOR_CODES[color]
            menu_item = Gtk.MenuItem(label=color_name, visible=True)
            menu_item.get_child().set_markup("<span foreground='%s'>\u25A6</span>  %s" % (color_code, color_name))
            menu_item.connect('activate', self.set_color, color)
            menu.append(menu_item)

        color_button.set_popup(menu)

        menu = Gtk.Menu()

        bold_item = Gtk.MenuItem(label=_("Bold"), visible=True)
        bold_item.get_child().set_markup("<b>%s</b>" % _("Bold"))
        bold_item.connect('activate', self.apply_format, 'bold')
        menu.append(bold_item)

        italic_item = Gtk.MenuItem(label=_("Italic"), visible=True)
        italic_item.get_child().set_markup("<i>%s</i>" % _("Italic"))
        italic_item.connect('activate', self.apply_format, 'italic')
        menu.append(italic_item)

        monospace_item = Gtk.MenuItem(label=_("Fixed Width"), visible=True)
        monospace_item.get_child().set_markup("<tt>%s</tt>" % _("Fixed Width"))
        monospace_item.connect('activate', self.apply_format, 'monospace')
        menu.append(monospace_item)

        underline_item = Gtk.MenuItem(label=_("Underline"), visible=True)
        underline_item.get_child().set_markup("<u>%s</u>" % _("Underline"))
        underline_item.connect('activate', self.apply_format, 'underline')
        menu.append(underline_item)

        strikethrough_item = Gtk.MenuItem(label=_("Strikethrough"), visible=True)
        strikethrough_item.get_child().set_markup("<s>%s</s>" % _("Strikethrough"))
        strikethrough_item.connect('activate', self.apply_format, 'strikethrough')
        menu.append(strikethrough_item)

        highlight_item = Gtk.MenuItem(label=_("Highlight"), visible=True)
        highlight_item.get_child().set_markup("<span background='yellow' foreground='black'>%s</span>" % _("Highlight"))
        highlight_item.connect('activate', self.apply_format, 'highlight')
        menu.append(highlight_item)

        header_item = Gtk.MenuItem(label=_("Header"), visible=True)
        header_item.get_child().set_markup("<span size='large'>%s</span>" % _("Header"))
        header_item.connect('activate', self.apply_format, 'header')
        menu.append(header_item)

        menu.append(Gtk.SeparatorMenuItem(visible=True))

        for (scale_id, scale_name, scale_value) in FONT_SCALES:
            font_scale_item = Gtk.MenuItem(label=scale_name, visible=True)
            font_scale_item.get_child().set_markup("<span size='%s'>%s</span>" % (scale_value, scale_name))
            font_scale_item.connect('activate', self.apply_format, scale_id)
            menu.append(font_scale_item)

        menu.append(Gtk.SeparatorMenuItem(visible=True))

        self.checklist_item = Gtk.MenuItem(label="\u25A2 %s" % _("Toggle Checklist"), visible=True)
        self.checklist_item.connect('activate', self.buffer.toggle_checklist)
        menu.append(self.checklist_item)

        self.bullet_item = Gtk.MenuItem(label="\u25CF %s" % _("Toggle Bullets"), visible=True)
        self.bullet_item.connect('activate', self.buffer.toggle_bullets)
        menu.append(self.bullet_item)

        text_button.set_popup(menu)

    def set_color(self, menu, color):
        if color == self.color:
            return

        self.get_style_context().remove_class(self.color)
        self.get_style_context().add_class(color)
        self.color = color

        self.emit('update')

    def set_font(self, *args):
        self.title_style_manager.set_from_pango_font_string(self.app.settings.get_string('font'))
        self.view_style_manager.set_from_pango_font_string(self.app.settings.get_string('font'))

    def apply_format(self, m, format_type):
        self.buffer.tag_selection(format_type)

    def remove(self, *args):
        # this is ugly but I'm not sure how to make it look better :)
        if (self.app.settings.get_boolean('disable-delete-confirm') or
            (not self.title.get_text() and self.buffer.get_char_count() == 0) or
            confirm(_("Delete Note"), _("Are you sure you want to remove this note?"),
                    self, self.app.settings, 'disable-delete-confirm')):
            self.emit('removed')
            self.destroy()

    def duplicate(self, *args):
        self.app.duplicate_note(self)

    def set_title(self, *args):
        self.title_text = self.title.get_text()
        self.title_box.remove(self.title)

        self.title = Gtk.Entry(text=self.title_text, visible=True, name='title')
        self.title_box.pack_start(self.title, False, False, 0)

        self.title.key_id = self.title.connect('key-press-event', self.save_title)
        self.title.focus_id = self.title.connect('focus-out-event', self.save_title)

        self.title_box.reorder_child(self.title, 0)
        self.title_hover.disable()

        self.title.grab_focus()

    def save_title(self, w, event):
        save = False
        if event.type == Gdk.EventType.FOCUS_CHANGE:
            save = True
        else:
            if event.keyval in (Gdk.KEY_Return, Gdk.KEY_ISO_Enter, Gdk.KEY_KP_Enter):
                save = True
            elif event.keyval != Gdk.KEY_Escape:
                return Gdk.EVENT_PROPAGATE

        self.title.disconnect(self.title.key_id)
        self.title.disconnect(self.title.focus_id)

        if save:
            self.title_text = self.title.get_text()

        self.view.grab_focus()

        self.title_box.remove(self.title)

        self.title = Gtk.Label(label=self.title_text, visible=True, name='title')
        self.title_box.pack_start(self.title, False, False, 0)

        self.title_box.reorder_child(self.title, 0)
        self.title_hover.enable()

        if save:
            self.emit('update')

        return Gdk.EVENT_STOP

class SettingsWindow(XApp.PreferencesWindow):
    def __init__(self, app):
        super(SettingsWindow, self).__init__(skip_taskbar_hint=False, title=_("Preferences"))

        # general settings
        page = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        page.pack_start(GSettingsSwitch(_("Show notes on all desktops"), SCHEMA, 'desktop-window-state'), False, False, 0)
        page.pack_start(GSettingsSwitch(_("Show in taskbar"), SCHEMA, 'show-in-taskbar'), False, False, 0)
        page.pack_start(GSettingsSwitch(_("Tray icon"), SCHEMA, 'show-in-tray'), False, False, 0)
        page.pack_start(GSettingsSwitch(_("Show the main window automatically"), SCHEMA, 'show-manager', dep_key=SCHEMA+'/show-in-tray'), False, False, 0)
        self.add_page(page, 'general', _("General"))

        # note related settings
        page = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        page.pack_start(GSettingsSpinButton(_("Default height"), SCHEMA, 'default-height', mini=50, maxi=2000, step=10), False, False, 0)
        page.pack_start(GSettingsSpinButton(_("Default width"), SCHEMA, 'default-width', mini=50, maxi=2000, step=10), False, False, 0)
        start_posititions = [(x, y) for x, y in START_POSITIONS.items()]
        page.pack_start(GSettingsComboBox(_("Default position"), SCHEMA, 'default-position', options=start_posititions, valtype=str), False, False, 0)
        try:
            colors = [(x, y) for x, y in COLORS.items()]
            colors.append(('sep', ''))
            colors.append(('cycle', _('Cycle Colors')))

            page.pack_start(GSettingsComboBox(_("Default color"), SCHEMA, 'default-color', options=colors, valtype=str, separator='sep'), False, False, 0)
        except Exception as e:
            colors = [(x, y) for x, y in COLORS.items()]
            colors.append(('cycle', _('Cycle Colors')))

            page.pack_start(GSettingsComboBox(_("Default color"), SCHEMA, 'default-color', options=colors, valtype=str), False, False, 0)

        page.pack_start(GSettingsFontButton(_("Font"), SCHEMA, 'font', level=Gtk.FontChooserLevel.SIZE), False, False, 0)
        page.pack_start(GSettingsSwitch(_("Show spelling mistakes"), SCHEMA, 'inline-spell-check'), False, False, 0)

        self.add_page(page, 'notes', _("Notes"))

        # backups
        page = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        page.pack_start(GSettingsSwitch(_("Automatic backups"), SCHEMA, 'automatic-backups'), False, False, 0)
        page.pack_start(GSettingsSpinButton(_("Time between backups"), SCHEMA, 'backup-interval', units=_("hours")), False, False, 0)
        obm_tooltip = _("Set this to zero if you wish to keep all backups indefinitely")
        page.pack_start(GSettingsSpinButton(_("Number to keep"), SCHEMA, 'old-backups-max', tooltip=obm_tooltip), False, False, 0)

        self.add_page(page, 'backup', _("Backups"))

        # autostart
        page = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        page.pack_start(GSettingsSwitch(_("Start automatically"), SCHEMA, 'autostart'), False, False, 0)
        page.pack_start(GSettingsSwitch(_("Show notes on the screen"), SCHEMA, 'autostart-notes-visible', dep_key=SCHEMA+'/autostart'), False, False, 0)
        self.add_page(page, 'autostart', _("Automatic start"))

        self.show_all()

class ShortcutsWindow(Gtk.ShortcutsWindow):
    def __init__(self):
        super(ShortcutsWindow, self).__init__()

        section = Gtk.ShortcutsSection(visible=True)

        for group, items in SHORTCUTS.items():
            group = Gtk.ShortcutsGroup(title=group, visible=False)
            section.add(group)

            for shortcut in items:
                shortcut_item = Gtk.ShortcutsShortcut(title=shortcut[0], accelerator=shortcut[1], visible=True)
                group.add(shortcut_item)

        group = Gtk.ShortcutsGroup(title=_("Text Size"), visible=False)
        section.add(group)
        for i in range(len(FONT_SCALES)):
            shortcut_item = Gtk.ShortcutsShortcut(title=FONT_SCALES[i][1], accelerator='<ctrl>%d' % (i + 2), visible=True)
            group.add(shortcut_item)

        self.add(section)
        self.show_all()

class Application(Gtk.Application):

    def __init__(self, *args):
        super(Application, self).__init__(*args, application_id=APPLICATION_ID, flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE)

        self.status_icon = None
        self.has_activated = False
        self.notes = []
        self.settings_window = None
        self.keyboard_shortcuts = None
        # There's no use creating the manager if a user is never going to use it, so we don't until it's asked for.
        # Therefore, we should never assume that the manager already exists, and handle the situation gracefully if it
        # is none.
        self.manager = None
        self.notes_hidden = False
        self.autostart_mode = False # indicates if we're in autostart mode
        self.dbus_register_id = 0
        self.group_buttons = {}

        self.add_main_option(
            'autostart',
            ord('a'),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            'Autostart mode',
            None)
        self.add_main_option(
            'new',
            ord('n'),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            'Create a new note',
            None)

        self.add_main_option(
            'toggle',
            ord('t'),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            'Toggle note visibility',
            None)

    def do_command_line(self, command_line):
        options = command_line.get_options_dict()
        options = options.end().unpack()

        if 'autostart' in options:
            settings = Gio.Settings(schema_id=SCHEMA)
            if not settings.get_boolean("autostart"):
                sys.exit()
            else:
                self.autostart_mode = True

        if not self.has_activated:
            self.activate()

        if 'new' in options:
            self.new_note(None)

        if 'toggle' in options:
            if self.notes_hidden:
                self.activate_notes(Gtk.get_current_event_time())
            else:
                self.hide_notes()

        return 0

    def do_activate(self):
        if self.has_activated:
            for note in self.notes:
                note.restore()
            self.open_manager()
            return

        Gtk.Application.do_activate(self)

        self.settings = Gio.Settings(schema_id=SCHEMA)

        self.dummy_window = Gtk.Window(title=_("Notes"), default_height=1, default_width=1, decorated=False, deletable=False, name='dummy-window')
        self.dummy_window.show()

        self.file_handler = FileHandler(self.settings, self.dummy_window)

        if self.settings.get_boolean('first-run'):
            self.first_run()

        # Backwards compatibility
        # - Update random color option to cycle
        if self.settings.get_string('default-color') == 'random':
            self.settings.set_string('default-color', 'cycle')

        self.file_handler.connect('lists-changed', self.on_lists_changed)
        self.group_update_id = self.file_handler.connect('group-changed', self.on_group_changed)
        self.file_handler.connect('group-name-changed', self.on_group_name_changed)
        self.file_handler.connect('saved', self.on_save)

        if self.settings.get_boolean('show-in-tray'):
            self.create_status_icon()

        self.settings.connect('changed::show-in-tray', self.update_tray_icon)
        self.settings.connect('changed::show-in-taskbar', self.update_dummy_window)
        self.active_group_changed_id = self.settings.connect('changed::active-group', self.on_active_group_changed)
        self.update_dummy_window()

        self.note_group = self.settings.get_string('active-group')
        group_names = self.file_handler.get_note_group_names()
        if self.note_group not in group_names:
            if len(group_names) > 0:
                self.note_group = group_names[0]
                self.settings.set_string('active-group', self.note_group)
            else:
                self.file_handler.new_group(self.note_group)

        provider = Gtk.CssProvider()
        provider.load_from_path(STYLE_SHEET_PATH)

        Gtk.StyleContext.add_provider_for_screen (Gdk.Screen.get_default(), provider, 600)

        self.load_notes()

        self.hold()

        if self.autostart_mode:
            self.autostart_mode = False
            if not self.settings.get_boolean("autostart-notes-visible"):
                self.hide_notes()
        else:
            if self.settings.get_boolean("show-manager"):
                self.open_manager()

        if not self.settings.get_boolean("show-in-tray"):
            self.open_manager()

        self.has_activated = True

    def do_dbus_register(self, connection, object_path):
        dbus_interface_info = Gio.DBusNodeInfo.new_for_xml(DBUS_INTERFACE_XML)
        self.dbus_register_id = connection.register_object(
            object_path, dbus_interface_info.interfaces[0], self.dbus_method_callback, None, None)

        return Gio.Application.do_dbus_register(self, connection, object_path)

    def do_dbus_unregister(self, connection, object_path):
        if self.dbus_register_id > 0:
            connection.unregister_object(self.dbus_register_id)
            self.dbus_register_id = 0


        Gio.Application.do_dbus_unregister(self, connection, object_path)

    def dbus_method_callback(self, connection, sender, path, iface_name, method_name, params, invocation, user_data=None):
        if method_name == 'ShowNotes':
            self.activate_notes(0)

        elif method_name == 'NewNote':
            x, y, direction = self.get_direction()
            x, y = self.find_note_location(x, y, direction)

            self.add_note({'text': params.unpack()[0], 'x': x, 'y': y})

        elif method_name == 'NewNoteBlank':
            self.new_note()

        elif method_name == 'ReloadNotesFromFile':
            self.reload_notes_from_file()

    def first_run(self):
        gnote_dir = os.path.join(GLib.get_user_data_dir(), 'gnote')

        if os.path.exists(gnote_dir):
            contents = os.listdir(gnote_dir)

            import_notes = []
            for file in contents:
                path = os.path.join(gnote_dir, file)
                if os.path.isfile(path) and path.endswith('.note'):
                    import_notes.append(path)

            if len(import_notes) > 0:
                resp = confirm(_("Notes"),
                               _("Would you like to import your notes from Gnote? This will not change your Gnote notes in any way."),
                               window=self.dummy_window)

                if resp:
                    coordinates = 40
                    for file in import_notes:
                        (group_name, info, is_template) = gnote_to_internal_format(file)
                        if not is_template:
                            info['color'] = 'yellow'
                            info['x'] = coordinates
                            info['y'] = coordinates

                            if group_name not in self.file_handler.get_note_group_names():
                                self.file_handler.new_group(group_name)

                            group_list = self.file_handler.get_note_list(group_name)
                            group_list.append(info)
                            self.file_handler.update_note_list(group_list, group_name)

                            coordinates += 20

        # Create a default group
        if len(self.file_handler.get_note_group_names()) == 0:
            self.file_handler.update_note_list([{'text':'', 'color':'yellow', 'x': 20, 'y': 20}], _("Group 1"))

        self.settings.set_boolean('first-run', False)

    def create_status_icon(self):
        self.status_icon = XApp.StatusIcon()
        self.status_icon.set_name('sticky')
        self.status_icon.set_icon_name('sticky-symbolic')
        self.status_icon.set_tooltip_text('%s\n<i>%s</i>\n<i>%s</i>' % (_("Notes"),
                                                                        _("Left click to toggle notes"),
                                                                        _("Middle click to toggle the manager")))
        self.status_icon.set_visible(True)
        self.status_icon.connect('button-press-event', self.on_tray_button_pressed)

        self.rebuild_tray_context_menu()

    def on_tray_button_pressed(self, icon, x, y , button, time, panel_position):
        if button == 1:
            self.activate_notes(time)
        elif button == 2:
            self.toggle_manager(time)

    def rebuild_tray_context_menu(self):
        if self.status_icon is None:
            return

        self.group_buttons = {}

        self.context_menu = Gtk.Menu()
        item = Gtk.MenuItem(label=_("New Note"))
        item.connect('activate', self.new_note)
        self.context_menu.append(item)

        item = Gtk.MenuItem(label=_("Manage Notes"))
        item.connect('activate', self.open_manager)
        self.context_menu.append(item)

        self.context_menu.append(Gtk.SeparatorMenuItem())

        last_item = None
        for group in self.file_handler.get_note_group_names():
            item = Gtk.RadioMenuItem(label=group)

            if last_item is not None:
                item.join_group(last_item)

            if group == self.settings.get_string('active-group'):
                item.set_active(True)

            last_item = item

            item.connect('activate', self.on_tray_group_selected, group)
            self.context_menu.append(item)
            self.group_buttons[group] = item

        self.context_menu.append(Gtk.SeparatorMenuItem())

        item = Gtk.MenuItem(label=_("Preferences"))
        item.connect('activate', self.open_settings_window)
        self.context_menu.append(item)

        self.context_menu.append(Gtk.SeparatorMenuItem())

        item = Gtk.MenuItem(label=_("Quit"))
        item.connect('activate', self.quit_app)
        self.context_menu.append(item)

        self.context_menu.show_all()

        self.status_icon.set_secondary_menu(self.context_menu)

    def destroy_status_icon(self):
        self.status_icon.set_visible(False)
        self.status_icon = None

    def update_tray_icon(self, *args):
        if self.settings.get_boolean('show-in-tray'):
            self.create_status_icon()
        else:
            self.open_manager()
            self.destroy_status_icon()

    def on_tray_group_selected(self, widget, name):
        self.settings.set_string('active-group', name)

    def on_active_group_changed(self, settings, key):
        group = settings.get_string('active-group')
        if self.status_icon and not self.group_buttons[group].get_active():
            self.settings.disconnect(self.active_group_changed_id)
            self.group_buttons[group].set_active(True)
            self.active_group_changed_id = self.settings.connect('changed::active-group', self.on_active_group_changed)

        self.change_visible_note_group()

    def update_dummy_window(self, *args):
        if self.settings.get_boolean('show-in-taskbar') and not self.notes_hidden:
            self.dummy_window.set_skip_taskbar_hint(False)
        else:
            self.dummy_window.set_skip_taskbar_hint(True)

        self.dummy_window.move(-2, -2)

        if self.settings.get_boolean('desktop-window-state'):
            self.dummy_window.stick()

    def activate_notes(self, time):
        for note in self.notes:
            if note.is_active():
                self.hide_notes()
                return

        self.dummy_window.present_with_time(time)

        for note in self.notes:
            note.restore(time)

        if len(self.notes) == 0:
            self.new_note()

        self.notes_hidden = False
        self.update_dummy_window()

    def hide_notes(self):
        for note in self.notes:
            note.hide()
        
        self.notes_hidden = True
        self.update_dummy_window()

    def find_note_location(self, x, y, direction):
        while True:
            found = False
            for note_info in self.file_handler.get_note_list(self.note_group):
                if note_info['x'] == x and note_info['y'] == y:
                    found = True
                    break

            if not found:
                break

            x += 20 * direction[0]
            y += 60 * direction[1]

        # make sure the note doesn't get placed off-screen or over a panel
        display = Gdk.Display.get_default()
        monitor = display.get_monitor_at_point(x, y)
        workarea = monitor.get_workarea()
        min_x = workarea.x
        max_x = workarea.x + workarea.width
        min_y = workarea.y
        max_y = workarea.y + workarea.height

        x_end = x + self.settings.get_uint('default-height')
        y_end = y + self.settings.get_uint('default-width')

        if x < min_x or x_end > max_x or y < min_y or y_end > max_y:
            # the calculated location wont work, so put the new note near the top left corner
            x = workarea.x + 20
            y = workarea.y + 20

        return x, y

    def get_direction(self, parent=None):
        monitor = Gdk.Display.get_primary_monitor(Gdk.Display.get_default())
        if monitor == None:
            # if user has no monitor configured, or uses Wayland, for now guess
            # monitor 0 (better than a crash)
            monitor = Gdk.Display.get_monitor(Gdk.Display.get_default(),0)

        workarea_rect = Gdk.Monitor.get_workarea(monitor)
        direction = [1, 1]
        if parent:
            if parent.x > (workarea_rect.width / 2):
                x = parent.x - 20
                direction[0] = -1
            else:
                x = parent.x + 20
            if parent.y > (workarea_rect.height / 2):
                y = parent.y - 60
                direction[1] = -1
            else:
                y = parent.y + 60
        else:
            pos_y, pos_x = self.settings.get_string('default-position').split('-')
            if pos_x == 'right':
                x = workarea_rect.width - self.settings.get_uint('default-width') - 20
                direction[0] = -1
            elif pos_x == 'center':
                x = workarea_rect.width / 2 - self.settings.get_uint('default-width') / 2
            else:
                x = 20
            if pos_y == 'bottom':
                y = workarea_rect.height - self.settings.get_uint('default-height') - 60
                direction[1] = -1
            elif pos_y == 'center':
                y = workarea_rect.height / 2 - self.settings.get_uint('default-height') / 2
            else:
                y = 20
        
        return (x, y, direction)

    def new_note(self, button=None, parent=None):
        x, y, direction = self.get_direction(parent)

        x, y = self.find_note_location(x, y, direction)

        info = {'x': x, 'y': y}

        self.add_note(info)

    def add_note(self, info):
        note = self.generate_note(info)
        note.present_with_time(Gtk.get_current_event_time())

        # Note is Gdk.WindowType.UTILITY - these don't get raised automatically
        # (see muffin: window.c:window_state_on_map)
        if not note.get_realized():
            note.realize()
        note.get_window().raise_()

        self.on_update()

    def generate_note(self, info={}):
        note = Note(self, self.dummy_window, info)
        note.connect('update', self.on_update)
        note.connect('removed', self.on_removed)

        self.notes.append(note)

        return note

    def load_notes(self):
        for note in self.notes:
            note.destroy()

        self.notes = []

        for note_info in self.file_handler.get_note_list(self.note_group):
            self.generate_note(note_info)

    def reload_notes_from_file(self):
        self.file_handler.load_notes()
        self.load_notes();

    def duplicate_note(self, new_note):
        new_note_info = new_note.get_info()
        new_note_info['x'] += 50
        new_note_info['y'] += 50

        self.add_note(new_note_info)

    def focus_note(self, note_info):
        for note in self.notes:
            if note.get_info() == note_info:
                note.present_with_time(0)

    def on_lists_changed(self, *args):
        if not self.note_group in self.file_handler.get_note_group_names():
            self.change_visible_note_group()
        else:
            self.load_notes()

        self.rebuild_tray_context_menu()

    def on_group_changed(self, f, group_name):
        if self.note_group == group_name:
            self.load_notes()

    def on_group_name_changed(self, f, old_name, new_name):
        if self.note_group == old_name:
            self.change_visible_note_group(new_name)

        if self.settings.get_string('active-group') == old_name:
            self.settings.set_string('active-group', new_name)

        self.rebuild_tray_context_menu()

    def on_save(self, *args):
        self.get_dbus_connection().emit_signal(None, DBUS_PATH, APPLICATION_ID, 'NotesChanged', None)

    def change_visible_note_group(self, group=None):
        default = self.settings.get_string('active-group')
        if group is None:
            self.note_group = default
        else:
            self.note_group = group

        group_names = self.file_handler.get_note_group_names()
        if self.note_group not in group_names:
            if len(group_names) > 0:
                self.note_group = group_names[0]
            else:
                self.file_handler.new_group(default)
                self.note_group = default

        self.load_notes()

    def open_manager(self, *args, time=0):
        if self.manager:
            if time == 0:
                time = Gtk.get_current_event_time()
            self.manager.window.present_with_time(time)
            return

        self.manager = NotesManager(self, self.file_handler)
        self.manager.window.connect('delete-event', self.manager_closed)

    def toggle_manager(self, time):
        if self.manager and self.manager.window.is_active() and self.manager.window.is_visible():
            self.manager.window.hide()
        else:
            self.open_manager(time=time)

    def manager_closed(self, *args):
        if self.status_icon is None:
            self.quit_app()

        self.manager.window.hide()

        return Gdk.EVENT_STOP

    def open_settings_window(self, *args):
        if self.settings_window:
            self.settings_window.present_with_time(Gtk.get_current_event_time())
            return

        self.settings_window = SettingsWindow(self)
        self.settings_window.connect('destroy', self.settings_window_closed)

        self.settings_window.show_all()

    def open_about(self, widget):
        dlg = Gtk.AboutDialog()
        dlg.set_transient_for(self.manager.window)
        dlg.set_modal(True)
        dlg.set_title(_("About"))
        dlg.set_program_name(_("Notes"))
        dlg.set_comments(_("Take notes and stay organized"))
        try:
            h = open('/usr/share/common-licenses/GPL', encoding="utf-8")
            s = h.readlines()
            gpl = ""
            for line in s:
                gpl += line
            h.close()
            dlg.set_license(gpl)
        except Exception as e:
            print (e)

        dlg.set_version("__DEB_VERSION__")
        dlg.set_icon_name("sticky")
        dlg.set_logo_icon_name("sticky")
        dlg.set_website("https://www.github.com/linuxmint/sticky")
        def close(w, res):
            if res == Gtk.ResponseType.CANCEL or res == Gtk.ResponseType.DELETE_EVENT:
                w.destroy()
        dlg.connect("response", close)
        dlg.show()

    def settings_window_closed(self, *args):
        self.settings_window = None

    def open_keyboard_shortcuts(self, *args):
        if self.keyboard_shortcuts:
            self.keyboard_shortcuts.present_with_time(Gtk.get_current_event_time())
            return

        self.keyboard_shortcuts = ShortcutsWindow()
        self.keyboard_shortcuts.connect('destroy', self.keyboard_shortcuts_closed)
        self.keyboard_shortcuts.show_all()

    def keyboard_shortcuts_closed(self, *args):
        self.keyboard_shortcuts = None

    def on_update(self, *args):
        info = []
        for note in self.notes:
            info.append(note.get_info())

        self.file_handler.handler_block(self.group_update_id)
        self.file_handler.update_note_list(info, self.note_group)
        self.file_handler.handler_unblock(self.group_update_id)

    def on_removed(self, note):
        self.notes.remove(note)
        self.on_update()

    def quit_app(self, *args):
        self.file_handler.flush()

        for note in self.notes:
            note.destroy()

        self.quit()

if __name__ == "__main__":
    sticky = Application()
    sticky.run(sys.argv)



#!/usr/bin/python3

import os
import json
import time
import re

from gi.repository import Gio, GLib, GObject, Gtk

CONFIG_DIR = os.path.join(GLib.get_user_config_dir(), 'sticky')
CONFIG_PATH = os.path.join(CONFIG_DIR, 'notes.json')
SAVE_DELAY = 3

backup_file_name = re.compile(r"\Abackup-[0-9]{10,}\.json$", re.IGNORECASE)

class FileHandler(GObject.Object):
    @GObject.Signal(flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(str,),
                    accumulator=GObject.signal_accumulator_true_handled)
    def group_changed(self, group_name):
        pass

    @GObject.Signal(flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(str, str),
                    accumulator=GObject.signal_accumulator_true_handled)
    def group_name_changed(self, old_name, new_name):
        pass

    @GObject.Signal(flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    accumulator=GObject.signal_accumulator_true_handled)
    def lists_changed(self):
        pass

    @GObject.Signal(flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    accumulator=GObject.signal_accumulator_true_handled)
    def saved(self):
        pass

    def __init__(self, settings, window):
        super(FileHandler, self).__init__()

        self.settings = settings
        self.window = window
        self.save_timer_id = 0
        self.backup_timer_id = 0
        self.notes_lists = {}

        if os.path.exists(CONFIG_PATH):
            self.load_notes()

        self.settings.connect('changed::automatic-backups', self.check_backup)
        self.settings.connect('changed::backup-interval', self.check_backup)
        self.check_backup()

    def load_notes(self, *args):
        with open(CONFIG_PATH, 'r') as file:
            info = json.loads(file.read())

        self.notes_lists = info

    def get_note_list(self, group_name):
        return self.notes_lists[group_name]

    def get_note_group_names(self):
        return list(self.notes_lists.keys())

    def update_note_list(self, notes_list, group_name):
        self.notes_lists[group_name] = notes_list

        self.queue_save()

        self.emit('group-changed', group_name)

    def queue_save(self):
        if self.save_timer_id > 0:
            GLib.source_remove(self.save_timer_id)

        self.save_timer_id = GLib.timeout_add_seconds(SAVE_DELAY, self.save_note_list)

    def save_to_file(self, file_path):
        with open(file_path, 'w+') as file:
            file.write(json.dumps(self.notes_lists, indent=4))

    def save_note_list(self):
        self.save_timer_id = 0

        if not os.path.exists(CONFIG_DIR):
            os.makedirs(CONFIG_DIR)

        self.save_to_file(CONFIG_PATH)
        self.emit('saved')

    def check_backup(self, *args):
        if self.backup_timer_id:
            GLib.source_remove(self.backup_timer_id)
            self.backup_timer_id = 0

        if not self.settings.get_boolean('automatic-backups'):
            return

        now = int(time.time())
        last_backup = self.settings.get_uint('latest-backup')
        interval = self.settings.get_uint('backup-interval')

        if last_backup == 0:
            # unless it was reset, this means the application was just started for the first time, so there's no point
            # in running a backup yet
            self.settings.set_uint('latest-backup', now)
            last_backup = now

        next_backup = last_backup + (interval * 3600)

        if next_backup < now:
            self.save_backup()
        else:
            self.backup_timer_id = GLib.timeout_add_seconds(next_backup - now, self.save_backup)

    def save_backup(self, *args):
        self.backup_timer_id = 0

        if not os.path.exists(CONFIG_DIR):
            os.makedirs(CONFIG_DIR)

        timestamp = int(time.time())
        path = os.path.join(CONFIG_DIR, 'backup-%d.json' % timestamp)
        self.save_to_file(path)

        self.settings.set_uint('latest-backup', timestamp)

        # remove old backups (if applicable)
        backups_keep = self.settings.get_uint('old-backups-max')
        if backups_keep > 0:
            backups = []
            for file in os.listdir(CONFIG_DIR):
                if backup_file_name.search(file):
                    backups.append(file)

            backups.sort()
            for file in backups[0:-backups_keep]:
                os.remove(os.path.join(CONFIG_DIR, file))

        self.check_backup()

    def delete_all_backups(self):
        for file in os.listdir(CONFIG_DIR):
            if backup_file_name.search(file):
                os.remove(os.path.join(CONFIG_DIR, file))

    def export_notes(self, menuitem, window):
        file_dialog = Gtk.FileChooserDialog(title=_("Export..."), action=Gtk.FileChooserAction.SAVE, transient_for=window)
        file_dialog.add_buttons(_("Cancel"), Gtk.ResponseType.CANCEL, _("Save"), Gtk.ResponseType.OK)
        file_dialog.set_current_folder(GLib.get_home_dir())
        file_dialog.set_current_name('notes.json')
        file_dialog.set_do_overwrite_confirmation(True)

        json_filter = Gtk.FileFilter()
        json_filter.set_name('JSON')
        json_filter.add_mime_type('application/json')
        file_dialog.add_filter(json_filter)

        text_filter = Gtk.FileFilter()
        text_filter.set_name(_("Plain Text"))
        text_filter.add_mime_type('text/plain')
        file_dialog.add_filter(text_filter)

        response = file_dialog.run()
        if response == Gtk.ResponseType.OK:
            file = file_dialog.get_filename()
            self.save_to_file(file)

        file_dialog.destroy()

    def import_notes(self, menuitem, window):
        file_dialog = Gtk.FileChooserDialog(title=_("Import..."), action=Gtk.FileChooserAction.OPEN, transient_for=self.window)
        file_dialog.add_buttons(_("Cancel"), Gtk.ResponseType.CANCEL, _("Open"), Gtk.ResponseType.OK)
        file_dialog.set_current_folder(GLib.get_home_dir())

        json_filter = Gtk.FileFilter()
        json_filter.set_name('JSON')
        json_filter.add_mime_type('application/json')
        file_dialog.add_filter(json_filter)

        text_filter = Gtk.FileFilter()
        text_filter.set_name(_("Plain Text"))
        text_filter.add_mime_type('text/plain')
        file_dialog.add_filter(text_filter)

        response = file_dialog.run()
        if response == Gtk.ResponseType.OK:
            file_path = file_dialog.get_filename()
            if file_path is not None:
                self.load_notes_from_path(file_path, window)

        file_dialog.destroy()

    def restore_backup(self, menuitem, window):
        dialog = Gtk.Dialog(title=_("Restore Backup"), transient_for=window)
        dialog.add_button(_("Cancel"), Gtk.ResponseType.CANCEL)
        dialog.add_button(_("Clear"), 20)
        restore_button = dialog.add_button(_("Restore"), Gtk.ResponseType.OK)
        dialog.set_default_response(Gtk.ResponseType.OK)

        content = dialog.get_content_area()

        backup_list = Gtk.ListBox(selection_mode=Gtk.SelectionMode.BROWSE)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_shadow_type(Gtk.ShadowType.IN)
        scrolled_window.set_min_content_height(200)
        scrolled_window.add(backup_list)

        content.pack_start(scrolled_window, True, True, 0)

        backups = []
        for file in os.listdir(CONFIG_DIR):
            if backup_file_name.search(file):
                backups.append(file)

        backups.sort()

        if len(backups) == 0:
            restore_button.set_sensitive(False)

        for file_name in backups:
            date = time.localtime(int(file_name[7:-5]))
            label = Gtk.Label(label=time.strftime('%c', date), margin=5)
            label.file = file_name
            backup_list.add(label)

        scrolled_window.show_all()

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            file_path = os.path.join(CONFIG_DIR, backup_list.get_selected_row().get_child().file)
            if file_path is not None:
                self.load_notes_from_path(file_path, window)
        elif response == 20:
            self.delete_all_backups()

        dialog.destroy()

    def load_notes_from_path(self, path, window):
        try:
            with open(path, 'r') as file:
                info = json.loads(file.read())

            # todo: needs validation here to ensure the file type is correct, and while we're at it, the validation
            # should really be added to load_notes() as well

            self.notes_lists = info
            self.save_note_list()

            self.emit('lists-changed')
        except Exception as e:
            message = Gtk.MessageDialog(text=_("Unable to restore: invalid or corrupted backup file"),
                                        buttons=Gtk.ButtonsType.CLOSE, transient_for=window)
            message.run()
            message.destroy()

    def flush(self):
        if self.save_timer_id > 0:
            GLib.source_remove(self.save_timer_id)

        self.save_note_list()

    def new_group(self, group_name):
        if group_name in self.notes_lists:
            if not confirm(_("Overwrite Existing Group"), _("There is already a group named '%s'. This action will overwrite it. Continue anyway?") % group_name, self.window):
                return False

        self.notes_lists[group_name] = []

        self.save_note_list()
        self.emit('lists-changed')

        return True

    def remove_group(self, group_name):
        if not confirm(_("Remove Group"), _("Are you sure you want to remove the group %s?") % group_name, self.window):
            return

        if group_name not in self.notes_lists:
            raise ValueError('invalid group name %s' % group_name)
        del self.notes_lists[group_name]

        self.save_note_list()
        self.emit('lists-changed')

    def change_group_name(self, old_group, new_group):
        self.notes_lists[new_group] = self.notes_lists.pop(old_group)

        self.save_note_list()
        self.emit('group-name-changed', old_group, new_group)

class HoverBox(Gtk.EventBox):
    def __init__(self, widget=None):
        super(HoverBox, self).__init__()
        if widget is not None:
            self.set_child_widget(widget)

        self.inhibited = False

        self.connect('enter-notify-event', self.set_child_widget_visibility)
        self.connect('leave-notify-event', self.set_child_widget_visibility)

    def set_child_widget(self, widget):
        self.child_widget = widget
        widget.set_no_show_all(True)
        widget.hide()

    def disable(self):
        self.inhibited = True
        self.set_child_widget_visibility()

    def enable(self):
        self.inhibited = False
        self.set_child_widget_visibility()

    def set_child_widget_visibility(self, *args):
        pointer_window = self.get_display().get_default_seat().get_pointer().get_window_at_position()[0]
        has_mouse = pointer_window == self.get_window() or pointer_window in self.get_window().get_children()

        if not self.inhibited and has_mouse:
            self.child_widget.show()
        else:
            self.child_widget.hide()

def prompt(title, message, window):
    dialog = Gtk.Dialog(title=title, transient_for=window, window_position=Gtk.WindowPosition.CENTER)
    dialog.add_button(_("Cancel"), Gtk.ResponseType.CANCEL)
    dialog.add_button(_("OK"), Gtk.ResponseType.OK)
    dialog.set_default_response(Gtk.ResponseType.OK)

    content = dialog.get_content_area()
    content.props.margin_left = 20
    content.props.margin_right = 20

    content.pack_start(Gtk.Label(label=message), False, False, 10)
    entry = Gtk.Entry(activates_default=True)
    content.pack_start(entry, False, False, 10)

    content.show_all()

    response = dialog.run()
    value = entry.get_text()

    dialog.destroy()

    return response == Gtk.ResponseType.OK, value

def confirm(title, message, window=None, settings=None, disable_key=None, disable_inverted=False):
    dialog = Gtk.Dialog(title=title, transient_for=window,
        window_position=(Gtk.WindowPosition.CENTER if window is None else Gtk.WindowPosition.CENTER_ON_PARENT))
    dialog.add_button(_("No"), Gtk.ResponseType.NO)
    dialog.add_button(_("Yes"), Gtk.ResponseType.YES)
    dialog.set_default_response(Gtk.ResponseType.YES)

    content = dialog.get_content_area()
    content.props.margin_left = 20
    content.props.margin_right = 20

    content.pack_start(Gtk.Label(label=message), False, False, 10)

    if disable_key is not None:
        disable_checkbox = Gtk.CheckButton.new_with_label(_("Don't ask me again"))
        content.pack_start(disable_checkbox, False, False, 0)
        flags = Gio.SettingsBindFlags.DEFAULT
        if disable_inverted:
            flags |= Gio.SettingsBindFlags.INVERT_BOOLEAN
        settings.bind(disable_key, disable_checkbox, 'active', flags)

    content.show_all()

    response = dialog.run()

    dialog.destroy()

    return response == Gtk.ResponseType.YES




#!/usr/bin/python3

import re
import xml.etree.ElementTree as etree

ip_number = r"(?:\d{1,2}|1\d{2}|2[0-4]\d|25[0-5])"
ip_address = r"(?:(?:" + ip_number + ".){3}" + ip_number + ")"
domain = r"(?:(?:[a-z\u00a1-\uffff0-9]-?)*[a-z\u00a1-\uffff0-9]+)(?:\.(?:[a-z\u00a1-\uffff0-9]-?)*[a-z\u00a1-\uffff0-9]+)*(?:\.(?:[a-z\u00a1-\uffff]{2,}))"

regex_string = r"(?:(?:https?|ftp)://)(?:\S+(?::\S*)?@)?(?:" + ip_address + r"|" + domain + r")(?::\d{2,5})?(?:/\S*)?(?:\?\S*)?$\Z"
regex = re.compile(regex_string, re.IGNORECASE)

def ends_with_url(string):
    return bool(regex.search(string))

def get_url_start(string):
    return regex.search(string)

# format conversion
GNOTE_TO_INTERNAL_MAP = {
    'bold': 'bold',
    'italic': 'italic',
    'underline': 'underline',
    'strikethrough': 'strikethrough',
    'highlight': 'highlight',
    'url': 'link',
    'small': 'small',
    'large': 'large',
    'huge': 'larger',
}

GNOTE_NS_PREFIX = '{http://beatniksoftware.com/tomboy}'

def gnote_to_internal_format(file_path):
    tree = etree.parse(file_path)
    root = tree.getroot()

    info = {}
    info['title'] = root.find(GNOTE_NS_PREFIX + 'title').text

    def process_element(element):
        text = ''

        tag_name = element.tag.split('}')[1]
        if tag_name in GNOTE_TO_INTERNAL_MAP:
            internal_tag = '#tag:%s:' % GNOTE_TO_INTERNAL_MAP[tag_name]
            text += internal_tag
        else:
            internal_tag = ''

        if element.text:
            text += element.text.replace('#', '##')

        for child in element:
            text += process_element(child)

        text += internal_tag

        if element.tail:
            text += element.tail.replace('#', '##')

        return text

    info['text'] = process_element(root.find(GNOTE_NS_PREFIX + 'text').find(GNOTE_NS_PREFIX + 'note-content'))

    category = None
    is_template = False
    try:
        tags = root.find(GNOTE_NS_PREFIX + 'tags')
        for tag in tags:
            if tag.text == "system:template":
                is_template = True
            elif tag.text.startswith('system:notebook:'):
                category = tag.text[16:]
    except Exception as e:
        pass

    if category is None:
        category = _("Unfiled")

    return category, info, is_template

def clean_text(text):
    current_index = 0
    new_text = ''
    while True:
        next_index = text.find('#', current_index)
        new_text += text[current_index:next_index]

        if next_index == -1:
            return new_text.lower()

        if text[next_index:next_index+2] == '##':
            new_text += '#'
            current_index = next_index + 2
        elif text[next_index:next_index+6] == '#check':
            current_index = next_index + 8
        elif text[next_index:next_index+7] == '#bullet':
            current_index = next_index + 8
        elif text[next_index:next_index+4] == '#tag':
            current_index = text.find(':', next_index+6) + 1
        else:
            current_index += 1



#!/usr/bin/python3

from gi.repository import Gdk, GLib, GObject, Gtk, Pango
from util import ends_with_url, get_url_start

TAG_DEFINITIONS = {
    'bold': {'weight': Pango.Weight.BOLD},
    'italic': {'style': Pango.Style.ITALIC},
    'monospace': {'family': 'Monospace'},
    'underline': {'underline': Pango.Underline.SINGLE},
    'strikethrough': {'strikethrough': True},
    'highlight': {'background': 'yellow'},
    'link': {'underline': Pango.Underline.SINGLE, 'foreground': 'blue'},
    'header': {'scale': 1.5, 'weight': 500, 'pixels-above-lines': 14, 'pixels-below-lines': 4},
    'small': {'scale': .8333333333333},
    'large': {'scale': 1.2},
    'larger': {'scale': 1.44}
}

FONT_SCALES = ['small', 'normal', 'large', 'larger']

LINK_ENDERS = ['\n', '\t', ' ', '.', ',', ';', ':']

class GenericAction(object):
    def maybe_join(self, new_action):
        return False

# Used whenever plain text is added to the buffer. Internal characters such as anchor points should be handled with
# ObjectInsertAction, etc.
class AdditionAction(GenericAction):
    def __init__(self, buffer, text, location):
        super(AdditionAction, self).__init__()
        self.buffer = buffer
        self.text = text

        self.position = location.get_offset()

    def undo(self):
        start = self.buffer.get_iter_at_offset(self.position)
        end = self.buffer.get_iter_at_offset(self.position + len(self.text))
        self.buffer.delete(start, end)

        # we generally want to put the cursor where the text was added/removed
        pos = self.buffer.get_iter_at_offset(self.position)
        self.buffer.select_range(pos, pos)

    def redo(self):
        self.buffer.insert(self.buffer.get_iter_at_offset(self.position), self.text)

        # we generally want to put the cursor at the end of the re-added text
        pos = self.buffer.get_iter_at_offset(self.position + len(self.text))
        self.buffer.select_range(pos, pos)

    def maybe_join(self, new_action):
        if not isinstance(new_action, AdditionAction):
            return False

        if new_action.text == '\n':
            return False

        if new_action.position == self.position + len(self.text):
            self.text += new_action.text
            return True

        return False

# Used whenever text is removed from the buffer.
class DeletionAction(GenericAction):
    def __init__(self, buffer, start, end):
        super(DeletionAction, self).__init__()
        self.buffer = buffer
        self.text = buffer.get_slice(start, end, True)

        self.position = start.get_offset()

        if buffer.get_has_selection():
            (buffer_start, buffer_end) = buffer.get_selection_bounds()
            if buffer_start.compare(start) == 0 and buffer_end.compare(end):
                self.deletion_type = 'selection'
            else:
                self.deletion_type = 'other'
        elif self.buffer.get_iter_at_mark(self.buffer.get_insert()).compare(end) == 0:
            self.deletion_type = 'backward'
        elif self.buffer.get_iter_at_mark(self.buffer.get_insert()).compare(start) == 0:
            self.deletion_type = 'forward'
        else:
            self.deletion_type = 'other'

    def undo(self):
        self.buffer.insert(self.buffer.get_iter_at_offset(self.position), self.text)

        # depending on the type of deletion, we want to put the cursor (and selection) in different places
        if self.deletion_type == 'forward':
            pos = self.buffer.get_iter_at_offset(self.position)
            self.buffer.select_range(pos, pos)
        elif self.deletion_type == 'backward':
            pos = self.buffer.get_iter_at_offset(self.position + len(self.text))
            self.buffer.select_range(pos, pos)
        else:
            # any other type of deletion, we probably want to just select the whole thing
            pos1 = self.buffer.get_iter_at_offset(self.position)
            pos2 = self.buffer.get_iter_at_offset(self.position + len(self.text))
            self.buffer.select_range(pos1, pos2)

    def redo(self):
        start = self.buffer.get_iter_at_offset(self.position)
        end = self.buffer.get_iter_at_offset(self.position + len(self.text))
        self.buffer.delete(start, end)

        # we usually want to put the cursor at the point of deletion
        pos = self.buffer.get_iter_at_offset(self.position)
        self.buffer.select_range(pos, pos)

    def maybe_join(self, new_action):
        if not isinstance(new_action, DeletionAction) or new_action.deletion_type != self.deletion_type:
            return False

        if self.deletion_type == 'forward' and new_action.position == self.position:
            self.text += new_action.text
            return True
        elif self.deletion_type == 'backward' and new_action.position == self.position - 1:
            self.text = new_action.text + self.text
            self.position = new_action.position
            return True

        return False

# Used for objects inserted at an anchor point such as checkbuttons, bullets, etc.
class ObjectInsertAction(GenericAction):
    def __init__(self, buffer, anchor, is_addition=True):
        super(ObjectInsertAction, self).__init__()
        self.buffer = buffer
        self.is_addition = is_addition
        if isinstance(anchor.get_widgets()[0], Gtk.CheckButton):
            self.anchor_type = 'check'
        elif isinstance(anchor.get_widgets()[0], Gtk.Image):
            self.anchor_type = 'bullet'

        self.position = buffer.get_iter_at_child_anchor(anchor).get_offset()

    def remove(self):
        start_anchor_iter = self.buffer.get_iter_at_offset(self.position)
        end_anchor_iter = self.buffer.get_iter_at_offset(self.position + 1)
        if self.anchor_type == 'check':
            self.checked = start_anchor_iter.get_child_anchor().get_widgets()[0].get_active()
        self.buffer.delete(start_anchor_iter, end_anchor_iter)

    def add(self):
        if self.anchor_type == 'check':
            self.buffer.add_check_button(self.buffer.get_iter_at_offset(self.position), checked=self.checked)
        elif self.anchor_type == 'bullet':
            self.buffer.add_bullet(self.buffer.get_iter_at_offset(self.position))

    def undo(self):
        if self.is_addition:
            self.remove()
        else:
            self.add()

    def redo(self):
        if self.is_addition:
            self.add()
        else:
            self.remove()

# Used for setting formatting tags
class TagAction(GenericAction):
    def __init__(self, buffer, name, start, end, is_addition=True):
        super(TagAction, self).__init__()
        self.buffer = buffer
        self.name = name
        self.is_addition = is_addition

        # there may be text between `start` and `end` that already has the tag applied (or doesn't, in the case of a
        # removal), and we need to find those ranges (if they exist) so we can restore that state properly when undoing
        current_iter = start.copy()
        in_tag = False
        range_start = None
        self.ranges = []
        while current_iter.compare(end) < 0:
            has_tag = current_iter.has_tag(buffer.get_tag_table().lookup(name))
            # if we're adding a tag, we only want to chage text that isn't already tagged
            # if it's a removal, we only want to change text that is already tagged
            if range_start is None and has_tag != is_addition:
                range_start = current_iter.get_offset()
            elif range_start and has_tag == is_addition:
                self.ranges.append((range_start, current_iter.get_offset()))
                range_start = None

            current_iter.forward_char()

        if range_start is not None:
            self.ranges.append((range_start, end.get_offset()))

    def remove(self):
        for (start, end) in self.ranges:
            self.buffer.remove_tag_by_name(self.name, self.buffer.get_iter_at_offset(start), self.buffer.get_iter_at_offset(end))

    def add(self):
        for (start, end) in self.ranges:
            self.buffer.apply_tag_by_name(self.name, self.buffer.get_iter_at_offset(start), self.buffer.get_iter_at_offset(end))

    def undo(self):
        if self.is_addition:
            self.remove()
        else:
            self.add()

    def redo(self):
        if self.is_addition:
            self.add()
        else:
            self.remove()

class ShiftAction(GenericAction):
    def __init__(self, buffer, start, end, is_up):
        super(ShiftAction, self).__init__()
        self.buffer = buffer
        self.is_up = is_up

        self.start = start.get_line()
        self.end = end.get_line()

    def shift_up(self):
        start_iter = self.buffer.get_iter_at_line(self.start)
        end_iter = self.buffer.get_iter_at_line(self.end)
        end_iter.forward_to_line_end()
        start_mark = self.buffer.create_mark(None, start_iter, False)
        end_mark = self.buffer.create_mark(None, end_iter, True)

        move_start = start_iter.copy()
        move_start.backward_line()
        move_end = move_start.copy()
        move_end.forward_to_line_end()

        move_start_mark = self.buffer.create_mark(None, move_start, False)
        move_end_mark = self.buffer.create_mark(None, move_end, True)

        self.buffer.insert_range(end_iter, move_start, move_end)
        anchor = self.buffer.get_iter_at_mark(move_start_mark).get_child_anchor()
        if anchor is not None:
            if isinstance(anchor.get_widgets()[0], Gtk.CheckButton):
                checked = anchor.get_widgets()[0].get_active()
                obj_action = self.buffer.add_check_button(self.buffer.get_iter_at_mark(end_mark), checked=checked)
            elif isinstance(anchor.get_widgets()[0], Gtk.Image):
                obj_action = self.buffer.add_bullet(self.buffer.get_iter_at_mark(end_mark))

        self.buffer.insert(self.buffer.get_iter_at_mark(end_mark), '\n', -1)
        delete_start = self.buffer.get_iter_at_mark(move_start_mark)
        delete_end = self.buffer.get_iter_at_mark(move_end_mark)
        delete_end.forward_char()
        self.buffer.delete(delete_start, delete_end)

        # since it gets really messy trying to restore the previous cursor position, just select the lines that are moving
        self.buffer.select_range(self.buffer.get_iter_at_mark(start_mark), self.buffer.get_iter_at_mark(end_mark))

        self.buffer.delete_mark(move_start_mark)
        self.buffer.delete_mark(move_end_mark)
        self.buffer.delete_mark(start_mark)
        self.buffer.delete_mark(end_mark)

        self.start -= 1
        self.end -= 1

    def shift_down(self):
        start_iter = self.buffer.get_iter_at_line(self.start)
        end_iter = self.buffer.get_iter_at_line(self.end)
        end_iter.forward_to_line_end()
        begin_mark = self.buffer.create_mark(None, start_iter, True)
        start_mark = self.buffer.create_mark(None, start_iter, False)
        end_mark = self.buffer.create_mark(None, end_iter, True)

        move_start = end_iter.copy()
        move_start.forward_line()
        move_end = move_start.copy()
        move_end.forward_to_line_end()

        move_start_mark = self.buffer.create_mark(None, move_start, False)
        move_end_mark = self.buffer.create_mark(None, move_end, True)

        self.buffer.insert_range(start_iter, move_start, move_end)
        self.buffer.insert(self.buffer.get_iter_at_mark(start_mark), '\n', -1)
        anchor = self.buffer.get_iter_at_mark(move_start_mark).get_child_anchor()
        if anchor is not None:
            if isinstance(anchor.get_widgets()[0], Gtk.CheckButton):
                checked = anchor.get_widgets()[0].get_active()
                obj_action = self.buffer.add_check_button(self.buffer.get_iter_at_mark(begin_mark), checked=checked)
            elif isinstance(anchor.get_widgets()[0], Gtk.Image):
                obj_action = self.buffer.add_bullet(self.buffer.get_iter_at_mark(begin_mark))

        delete_start = self.buffer.get_iter_at_mark(move_start_mark)
        delete_start.backward_char()
        delete_end = self.buffer.get_iter_at_mark(move_end_mark)
        self.buffer.delete(delete_start, delete_end)

        # since it gets really messy trying to restore the previous cursor position, just select the lines that are moving
        self.buffer.select_range(self.buffer.get_iter_at_mark(start_mark), self.buffer.get_iter_at_mark(end_mark))

        self.buffer.delete_mark(move_start_mark)
        self.buffer.delete_mark(move_end_mark)
        self.buffer.delete_mark(start_mark)
        self.buffer.delete_mark(end_mark)
        self.buffer.delete_mark(begin_mark)

        self.start += 1
        self.end += 1

    def undo(self):
        if self.is_up:
            self.shift_down()
        else:
            self.shift_up()

    def redo(self):
        if self.is_up:
            self.shift_up()
        else:
            self.shift_down()

# Used to combine multiple actions into one single undoable action. Actions should be passed in the same order in which
# they were performed. Failure to do so could result in order getting mixed up in the buffer.
class CompositeAction(GenericAction):
    def __init__(self, *args):
        super(CompositeAction, self).__init__()
        self.child_actions = args

    def undo(self):
        for action in reversed(self.child_actions):
            action.undo()

    def redo(self):
        for action in self.child_actions:
            action.redo()

# we need to subclass the check button to add a Gdk.Window so that we can change the cursor
class CheckBox(Gtk.CheckButton):
    def __init__(self, **kwargs):
        super(CheckBox, self).__init__(**kwargs)

        self.event_window = None

        self.set_has_window(False)

    def do_realize(self):
        Gtk.CheckButton.do_realize(self)

        parent_window = self.get_parent_window()
        self.set_window(parent_window)

        allocation = self.get_allocation()
        attributes = Gdk.WindowAttr()
        attributes.cursor = Gdk.Cursor.new_from_name(self.get_display(), 'default')
        attributes.window_type = Gdk.WindowType.CHILD
        attributes.x = allocation.x
        attributes.y = allocation.y
        attributes.width = allocation.width
        attributes.height = allocation.height
        attributes.wclass = Gdk.WindowWindowClass.INPUT_ONLY
        attributes.event_mask = self.get_events() | Gdk.EventMask.BUTTON_PRESS_MASK | Gdk.EventMask.BUTTON_RELEASE_MASK | Gdk.EventMask.TOUCH_MASK | Gdk.EventMask.ENTER_NOTIFY_MASK | Gdk.EventMask.LEAVE_NOTIFY_MASK | Gdk.EventMask.POINTER_MOTION_MASK

        attributes_mask = Gdk.WindowAttributesType.X | Gdk.WindowAttributesType.Y | Gdk.WindowAttributesType.CURSOR

        self.event_window = Gdk.Window.new(parent_window, attributes, attributes_mask)
        self.register_window(self.event_window)

    def do_map(self):
        Gtk.CheckButton.do_map(self)

        self.event_window.show()

    def do_unmap(self):
        Gtk.CheckButton.do_unmap(self)

        self.event_window.hide()

    def do_size_allocate(self, allocation):
        Gtk.CheckButton.do_size_allocate(self, allocation)

        if self.event_window:
            self.event_window.move_resize(allocation.x, allocation.y, allocation.width, allocation.height)

class NoteBuffer(Gtk.TextBuffer):
    # These values should not be modified directly.
    # internal_action_count keeps the "content-changed" signal from firing while the buffer performs several actions. It
    # should not be modified directly. Instead use
    #       with self.internal_action():
    #           do_something()
    internal_action_count = 0

    # in_composite and composite_actions will rarely be used in practice as it is generally much easier and
    # straightforward to construct the composite action directly as you perform them. This functionality is primarily
    # only for functions internal to the view and buffer.
    in_composite = 0
    composite_actions = []

    # used to keep track of undo and redo actions. Use self.add_undo_action() when creating a new action
    undo_actions = []
    redo_actions = []

    # Used to keep track of tags that are toggled when no text is selected. We need this so we can apply the tag when
    # typing.
    tag_toggles = []

    @GObject.Property
    def can_undo(self):
        return len(self.undo_actions)

    @GObject.Property
    def can_redo(self):
        return len(self.redo_actions)

    @GObject.Signal(flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    accumulator=GObject.signal_accumulator_true_handled)
    def content_changed(self):
        pass

    def __init__(self):
        super(NoteBuffer, self).__init__()

        self.tags = []

        for name, attributes in TAG_DEFINITIONS.items():
            self.tags.append(self.create_tag(name, **attributes))

        self.connect('delete-range', self.on_delete)
        self.connect('begin-user-action', self.begin_composite_action)
        self.connect('end-user-action', self.end_composite_action)
        self.connect('mark-set', self.on_mark_set)

    def set_view(self, view):
        def track_motion(v, event):
            if self.view.get_modifier_mask(Gdk.ModifierIntent.EXTEND_SELECTION):
                return Gdk.EVENT_PROPAGATE

            mouse_iter = self.view.get_iter_at_location(*self.view.window_to_buffer_coords(Gtk.TextWindowType.TEXT, event.x, event.y))[1]
            if mouse_iter.has_tag(self.get_tag_table().lookup('link')):
                self.view.props.window.set_cursor(Gdk.Cursor.new_from_name(Gdk.Display.get_default(), 'pointer'))

                return Gdk.EVENT_STOP

            return Gdk.EVENT_PROPAGATE

        def handle_click(v, event):
            if not(event.state & Gdk.ModifierType.CONTROL_MASK) or event.button != 1:
                return Gdk.EVENT_PROPAGATE

            tag = self.get_tag_table().lookup('link')
            mouse_iter = self.view.get_iter_at_location(*self.view.window_to_buffer_coords(Gtk.TextWindowType.TEXT, event.x, event.y))[1]
            if not mouse_iter.has_tag(tag):
                return Gdk.EVENT_PROPAGATE

            start_link = mouse_iter.copy()
            start_link.backward_to_tag_toggle(tag)
            end_link = mouse_iter.copy()
            end_link.forward_to_tag_toggle(tag)

            url = self.get_slice(start_link, end_link, False)
            Gtk.show_uri(None, url, event.time)

            return Gdk.EVENT_STOP

        self.view = view
        self.view.connect('motion-notify-event', track_motion)
        self.view.connect('button-press-event', handle_click)

    def trigger_changed(self, *args):
        if self.internal_action_count == 0:
            self.emit('content-changed')

    def internal_action(self, trigger_on_complete=True):
        class InternalActionHandler(object):
            def __enter__(a):
                self.internal_action_count += 1

            def __exit__(a, exc_type, exc_value, traceback):
                self.internal_action_count -= 1
                if self.internal_action_count == 0 and trigger_on_complete:
                    GLib.idle_add(self.trigger_changed)

        return InternalActionHandler()

    def get_internal_markup(self):
        on_tags = []
        off_tags = self.tags.copy()
        text = ''

        current_iter = self.get_iter_at_offset(0)
        while True:
            # first we close any open tags that don't continue on to the current character
            for tag in on_tags[:]:
                if not current_iter.has_tag(tag):
                    text += '#tag:%s:' % tag.props.name
                    off_tags.append(tag)
                    on_tags.remove(tag)

            # next we open any tags that start with the current character
            for tag in off_tags[:]:
                if current_iter.has_tag(tag):
                    text += '#tag:%s:' % tag.props.name
                    on_tags.append(tag)
                    off_tags.remove(tag)

            current_char = current_iter.get_char()
            if current_char == '#':
                # we need to escape '#' characters to avoid misinterpretation when we parse it later
                text += '##'
            elif current_iter.get_child_anchor() is not None:
                # object insertions (bullets and checkboxes)
                anchor_child = current_iter.get_child_anchor().get_widgets()[0]
                if isinstance(anchor_child, Gtk.CheckButton):
                    checked = anchor_child.get_active()
                    text += '#check:' + str(int(checked))
                elif isinstance(anchor_child, Gtk.Image):
                    text += '#bullet:'
            else:
                text += current_char

            if not current_iter.forward_char():
                break

        # If there are any open tags at this point, it means they go to the end of the text, so close them
        for tag in on_tags:
            text += '#tag:%s:' % tag.props.name

        return text

    def set_from_internal_markup(self, text):
        with self.internal_action(False):
            self.set_text('')

            current_index = 0
            open_tags = {}
            while True:
                next_index = text.find('#', current_index)
                if next_index == -1:
                    self.insert(self.get_end_iter(), text[current_index:])
                    break

                self.insert(self.get_end_iter(), text[current_index:next_index])

                if text[next_index:next_index+2] == '##':
                    self.insert(self.get_end_iter(), '#')
                    current_index = next_index + 2
                elif text[next_index:next_index+6] == '#check':
                    checked = bool(int(text[next_index+7]))
                    self.add_check_button(self.get_end_iter(), checked=checked)
                    current_index = next_index + 8
                elif text[next_index:next_index+7] == '#bullet':
                    self.add_bullet(self.get_end_iter())
                    current_index = next_index + 8
                elif text[next_index:next_index+4] == '#tag':
                    end_tag_index = text.find(':', next_index+6)
                    tag_name = text[next_index+5:end_tag_index]

                    if tag_name in open_tags:
                        mark = open_tags.pop(tag_name)
                        start = self.get_iter_at_mark(mark)
                        end = self.get_end_iter()
                        self.apply_tag_by_name(tag_name, start, end)
                        self.delete_mark(mark)
                    else:
                        open_tags[tag_name] = self.create_mark(None, self.get_end_iter(), True)

                    current_index = next_index + 6 + len(tag_name)
                else:
                    print('formatting error detected - attempting to fix')
                    self.insert(self.get_end_iter(), '#')
                    current_index = next_index + 1

    def undo(self, *args):
        if len(self.undo_actions) == 0:
            return

        with self.internal_action():
            action = self.undo_actions.pop()
            action.undo()
            self.redo_actions.append(action)

    def redo(self, *args):
        if len(self.redo_actions) == 0:
            return

        with self.internal_action():
            action = self.redo_actions.pop()
            action.redo()
            self.undo_actions.append(action)

    def begin_composite_action(self, *args):
        self.in_composite += 1

    def end_composite_action(self, *args):
        self.in_composite -= 1

        # if there are no actions that happen during the composite, there's nothing we need to do
        if self.in_composite or len(self.composite_actions) == 0:
            return

        # some times we get actions tagged as composite when they really shouldn't be, so if there's just one action
        # we don't want to put it inside a composite action as that will break joining actions if applicable
        if len(self.composite_actions) == 1:
            self.add_undo_action(self.composite_actions[0])
        else:
            self.add_undo_action(CompositeAction(*self.composite_actions))

        self.composite_actions.clear()

    def add_undo_action(self, action):
        if self.in_composite:
            self.composite_actions.append(action)
        else:
            self.undo_actions.append(action)
            self.redo_actions.clear()

    def do_insert_text(self, location, text, length):
        position = location.get_offset()

        action = AdditionAction(self, text, location)
        Gtk.TextBuffer.do_insert_text(self, location, text, length)

        if self.internal_action_count:
            return

        with self.internal_action():
            if text == '\n':
                # if the previous line starts with a checkbox or bullet, repeat
                next_line = self.get_iter_at_offset(position+1)
                prev_line = next_line.copy()
                prev_line.backward_line()
                anchor = prev_line.get_child_anchor()

                if anchor is not None:
                    if isinstance(anchor.get_widgets()[0], Gtk.CheckButton):
                        obj_action = self.add_check_button(next_line)
                    elif isinstance(anchor.get_widgets()[0], Gtk.Image):
                        obj_action = self.add_bullet(next_line)

                    action = CompositeAction(action, obj_action)

                    # `location` becomes invalidated by adding a check/bullet, but it is expected to be there when the
                    # function ends, so we need to reinitialize it. The default handler places it at the end of the
                    # inserted text, so that's what we'll do too
                    location.assign(self.get_iter_at_offset(position + 2))

            # 'location' has usually been shifted by this point, and since it may be used by something else later that
            # expects the new location, we need to use a different iter
            start = self.get_iter_at_offset(position)

            next_char = start.copy()
            next_char.forward_char()

            # check tag toggles
            if len(text) == 1:
                font_scale_changed = False
                for tag_name in self.tag_toggles:
                    # if the tag is a font scale, remove all other font scales
                    if tag_name in FONT_SCALES:
                        font_scale_changed = True
                        for scale_tag_name in FONT_SCALES:
                            if scale_tag_name != 'normal':
                                action = CompositeAction(action, self.strip_tag(scale_tag_name, start, next_char))

                    prev = start.copy()
                    if tag_name == 'normal':
                        # the 'normal' tag doesn't really exist but we still use it for removing other font scale tags
                        pass
                    elif prev.backward_char() & prev.has_tag(self.get_tag_table().lookup(tag_name)):
                        action = CompositeAction(action, self.strip_tag(tag_name, start, next_char))
                    else:
                        action = CompositeAction(action, self.add_tag(tag_name, start, next_char))

                for tag in start.get_toggled_tags(False):
                    tag_name = tag.props.name
                    if tag_name is None:
                        continue

                    # don't continue header tag onto new line
                    if text == '\n' and tag_name == 'header':
                        continue

                    # don't continue links across link-ending characters
                    if tag_name == 'link' and text in LINK_ENDERS:
                        continue

                    # don't continue the tag if it was toggled off
                    # also don't continue if it's a font scale tag and another was just applied
                    if tag_name not in self.tag_toggles and (tag_name not in FONT_SCALES or not font_scale_changed):
                        action = CompositeAction(action, self.add_tag(tag_name, start, next_char))

                self.tag_toggles = []

            if not self.props.can_undo or not self.undo_actions[-1].maybe_join(action):
                self.add_undo_action(action)

        # check for a link if text is a link-ending character
        if text in LINK_ENDERS:
            pre_text = self.get_slice(self.get_start_iter(), start, True)
            match = get_url_start(pre_text)
            if match:
                self.add_undo_action(self.add_tag('link', self.get_iter_at_offset(match.start()), start))

    def on_delete(self, buffer, start, end):
        if self.internal_action_count:
            return Gdk.EVENT_PROPAGATE

        # if there were tags, bullets or checkboxes here, we need to handle those first so that we can undo later
        actions = []
        with self.internal_action():
            current_iter = start.copy()

            # if there's a bullet or checkbox for the next 'character', remove that too
            if end.get_child_anchor() is not None:
                end.forward_char()

            start_mark = Gtk.TextMark()
            end_mark = Gtk.TextMark()

            self.add_mark(start_mark, start)
            self.add_mark(end_mark, end)

            open_tags = {}
            while current_iter.compare(end) < 0:
                anchor = current_iter.get_child_anchor()
                if anchor is not None:
                    current_offset = current_iter.get_offset()
                    action = ObjectInsertAction(self, anchor, is_addition=False)
                    actions.append(action)
                    action.remove()

                    current_iter = self.get_iter_at_offset(current_offset)
                    end.assign(self.get_iter_at_mark(end_mark))
                    start.assign(self.get_iter_at_mark(start_mark))

                for tag in current_iter.get_toggled_tags(True):
                    # ignore tags that don't have one of our names (i.e. spell checker)
                    if tag.props.name in TAG_DEFINITIONS:
                        open_tags[tag.props.name] = current_iter.get_offset()

                for tag in current_iter.get_toggled_tags(False):
                    # ignore tags that don't have one of our names (i.e. spell checker)
                    if tag.props.name not in TAG_DEFINITIONS:
                        continue

                    if tag.props.name in open_tags:
                        actions.append(TagAction(self, tag.props.name, self.get_iter_at_offset(open_tags[tag.props.name]), current_iter, False))
                        del open_tags[tag.props.name]
                    else:
                        actions.append(TagAction(self, tag.props.name, start, current_iter, False))

                current_iter.forward_char()

            for name, offset in open_tags.items():
                actions.append(TagAction(self, name, self.get_iter_at_offset(offset), end, False))

            self.delete_mark(start_mark)
            self.delete_mark(end_mark)

            # if it's just an object deletion, there's nothing left to remove, so there's no need to create an undo action
            if start.compare(end) != 0:
                actions.append(DeletionAction(self, start, end))

            if len(actions) == 0:
                return Gdk.EVENT_STOP
            elif len(actions) == 1:
                action = actions[0]
            else:
                action = CompositeAction(*actions)

            if self.props.can_undo and self.undo_actions[-1].maybe_join(action):
                return Gdk.EVENT_PROPAGATE

            self.add_undo_action(action)

    def tag_selection(self, tag_name):
        if self.get_has_selection():
            (start, end) = self.get_selection_bounds()

            current_iter = start.copy()
            remove_tag = True
            if tag_name in FONT_SCALES:
                remove_tag = False
            else:
                tag = self.get_tag_table().lookup(tag_name)
                while current_iter.compare(end) < 0:
                    if not current_iter.has_tag(tag):
                        remove_tag = False
                        break

                    current_iter.forward_char()

            # remove the tag if the whole selection already has it (unless it's a font scale tag)
            if remove_tag:
                self.add_undo_action(self.strip_tag(tag_name, start, end))
            else:
                self.add_undo_action(self.add_tag(tag_name, start, end))
        else:
            # if it's a font scale, we want to remove any existing font scale first
            # this has the side effect that font scales don't toggle like everything else, which is exactly what we want
            if tag_name in FONT_SCALES:
                for name in self.tag_toggles:
                    if name in FONT_SCALES:
                        self.tag_toggles.remove(name)

            if tag_name in self.tag_toggles:
                self.tag_toggles.remove(tag_name)
            else:
                self.tag_toggles.append(tag_name)

    def add_tag(self, tag_name, start, end):
        actions = []
        if tag_name in FONT_SCALES:
            for font_tag_name in FONT_SCALES:
                if font_tag_name == tag_name or font_tag_name == 'normal':
                    continue

                action = self.strip_tag(font_tag_name, start, end)
                if len(action.ranges) > 0:
                    actions.append(action)

        if tag_name != 'normal':
            actions.append(TagAction(self, tag_name, start, end))
            self.apply_tag_by_name(tag_name, start, end)
        self.trigger_changed()
        if len(actions) == 1:
            return actions[0]
        else:
            return CompositeAction(*actions)

    def strip_tag(self, tag_name, start, end):
        action = TagAction(self, tag_name, start, end, False)
        self.remove_tag_by_name(tag_name, start, end)
        self.trigger_changed()

        return action

    def add_check_button(self, a_iter, checked=False):
        with self.internal_action():
            anchor = self.create_child_anchor(a_iter)
            check_button = CheckBox(visible=True, active=checked, margin_right=5, margin_top=5)
            check_button.connect('toggled', self.trigger_changed)
            self.view.add_child_at_anchor(check_button, anchor)

            return ObjectInsertAction(self, anchor)

    def add_bullet(self, a_iter):
        with self.internal_action():
            anchor = self.create_child_anchor(a_iter)
            bullet = Gtk.Image(visible=True, icon_name='menu-bullet', pixel_size=16)
            self.view.add_child_at_anchor(bullet, anchor)

            return ObjectInsertAction(self, anchor)

    def toggle_checklist(self, *args):
        actions = []
        with self.internal_action():
            if self.get_has_selection():
                (start, end) = self.get_selection_bounds()
            else:
                start = end = self.get_iter_at_mark(self.get_insert())

            line_index_start = start.get_line()
            line_index_end = end.get_line()

            all_have_checkboxes = True
            for line in range(line_index_start, line_index_end + 1):
                if self.get_iter_at_line(line).get_child_anchor() is None:
                    all_have_checkboxes = False
                    break

            for line in range(line_index_start, line_index_end + 1):
                if all_have_checkboxes:
                    anchor = self.get_iter_at_line(line).get_child_anchor()
                    if anchor is not None:
                        action = ObjectInsertAction(self, anchor, False)
                        action.remove()
                        actions.append(action)
                else:
                    actions.append(self.add_check_button(self.get_iter_at_line(line)))

            if len(actions):
                self.add_undo_action(CompositeAction(*actions))

    def toggle_bullets(self, *args):
        actions = []
        with self.internal_action():
            if self.get_has_selection():
                (start, end) = self.get_selection_bounds()
            else:
                start = end = self.get_iter_at_mark(self.get_insert())

            line_index_start = start.get_line()
            line_index_end = end.get_line()

            all_have_bullets = True
            for line in range(line_index_start, line_index_end + 1):
                if self.get_iter_at_line(line).get_child_anchor() is None:
                    all_have_bullets = False
                    break

            for line in range(line_index_start, line_index_end + 1):
                if all_have_bullets:
                    anchor = self.get_iter_at_line(line).get_child_anchor()
                    if anchor is not None:
                        action = ObjectInsertAction(self, anchor, False)
                        action.remove()
                        actions.append(action)
                else:
                    actions.append(self.add_bullet(self.get_iter_at_line(line)))

        if len(actions):
            self.add_undo_action(CompositeAction(*actions))

    def on_return(self):
        if self.get_has_selection():
            return Gdk.EVENT_PROPAGATE

        cursor = self.get_iter_at_mark(self.get_insert())
        prev_char = cursor.copy()
        prev_char.backward_char()

        if not cursor.ends_line() or not prev_char.starts_line():
            return Gdk.EVENT_PROPAGATE

        anchor = prev_char.get_child_anchor()

        if anchor is not None:
            self.delete(prev_char, cursor)
            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def shift(self, is_up):
        if self.get_has_selection():
            (start, end) = self.get_selection_bounds()
        else:
            start = self.get_iter_at_mark(self.get_insert())
            end = start.copy()

        action = ShiftAction(self, start, end, is_up)

        with self.internal_action():
            if is_up:
                if start.compare(self.get_start_iter()) == 0:
                    return

                action.shift_up()

            else:
                if end.compare(self.get_end_iter()) == 0:
                    return

                action.shift_down()

        self.add_undo_action(action)

    def on_mark_set(self, b, l, mark):
        if mark.get_name() == 'insert':
            self.tag_toggles = []

    def test(self):
        print(ends_with_url(self.get_text(*self.get_bounds(), False)))



#!/usr/bin/python3

from gi.repository import Gdk, Gio, GLib, GObject, Gtk, Pango, XApp
from note_buffer import NoteBuffer
from common import HoverBox
from util import clean_text

NOTE_TARGETS = [Gtk.TargetEntry.new('note-entry', Gtk.TargetFlags.SAME_APP, 1)]

class NoteEntry(Gtk.Container):
    initialized = False

    def __init__(self, item, settings):
        super(NoteEntry, self).__init__(height_request=150,
                                        width_request=150,
                                        valign=Gtk.Align.START,
                                        halign=Gtk.Align.CENTER)

        self.item = item
        self.settings = settings

        self.set_has_window(False)
        self.style_manager = XApp.StyleManager(widget=self)

        self.title_bar = Gtk.Box(name='title-bar', visible=True)
        self.title_bar.pack_start(Gtk.Label(label=item.title, name='title', visible=True, margin_top=5, margin_bottom=5, ellipsize=Pango.EllipsizeMode.END), False, False, 0)
        self.title_bar.set_parent(self)

        self.buffer = NoteBuffer()
        self.text = Gtk.TextView(wrap_mode=Gtk.WrapMode.WORD_CHAR, populate_all=True, buffer=self.buffer, visible=True, sensitive=False)
        self.text.set_parent(self)

        self.buffer.set_view(self.text)
        self.buffer.set_from_internal_markup(item.text)

        self.settings.connect('changed::font', self.set_font)
        self.set_font()

        self.set_tooltip_text(item.title)

        self.initialized = True

        self.show_all()

    def do_size_allocate(self, allocation):
        Gtk.Widget.do_size_allocate(self, allocation)

        title_bar_height = self.title_bar.get_preferred_height()[1]

        title_rect = Gdk.Rectangle()
        title_rect.x = allocation.x
        title_rect.y = allocation.y
        title_rect.width = allocation.width
        title_rect.height = title_bar_height
        self.title_bar.size_allocate(title_rect)

        content_rect = Gdk.Rectangle()
        content_rect.x = allocation.x + 10
        content_rect.y = allocation.y + title_bar_height + 10
        content_rect.width = allocation.width - 20
        content_rect.height = allocation.height - title_bar_height - 20
        self.text.size_allocate(content_rect)

        self.set_clip(allocation)

    def do_get_preferred_height(self):
        return 150, 150

    def do_get_preferred_height_for_width(self, width):
        return 150, 150

    def do_get_preferred_width(self):
        return 150, 150

    def do_get_preferred_width_for_height(self, height):
        return 150, 150

    def do_destroy(self):
        if not self.initialized:
            return

        self.text.unparent()

        self.initialized = False
        Gtk.Container.do_destroy(self)

    def do_forall(self, include_internals, callback, *args):
        if include_internals:
            callback(self.title_bar, *args)
            callback(self.text, *args)

    def set_font(self, *args):
        self.style_manager.set_from_pango_font_string(self.settings.get_string('font'))

class GroupEntry(Gtk.ListBoxRow):
    def __init__(self, item):
        super(GroupEntry, self).__init__()
        self.item = item
        self.file_handler = self.item.file_handler

        self.props.height_request = 35

        self.hoverbox = HoverBox()
        self.add(self.hoverbox)

        self.menu = Gtk.Menu()

        self.new_item = Gtk.MenuItem(label=_("New"), visible=True)
        self.menu.append(self.new_item)

        self.menu.append(Gtk.SeparatorMenuItem(visible=True))

        item = Gtk.MenuItem(label=_("Edit"), visible=True)
        item.connect('activate', self.edit_group_name)
        self.menu.append(item)

        self.remove_item = Gtk.MenuItem(label=_("Remove"), visible=True)
        self.remove_item.connect('activate', self.remove_group)
        self.menu.append(self.remove_item)

        self.connect('popup-menu', self.on_popup)
        self.connect('button-press-event', self.on_button_press)
        self.connect('key-press-event', self.on_key_press)

        self.generate_content()

    def on_popup(self, *args):
        self.menu.popup_at_widget(self, Gdk.Gravity.CENTER, Gdk.Gravity.CENTER, None)

    def on_button_press(self, w, event):
        if event.button == 3:
            self.menu.popup_at_pointer(event)

    def on_key_press(self, w, event):
        if event.keyval == Gdk.KEY_Delete:
            self.remove_group()

            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def generate_content(self):
        self.box = Gtk.Box()
        self.hoverbox.add(self.box)

        label = Gtk.Label(label=self.item.name, halign=Gtk.Align.START, margin_start=5)
        self.box.pack_start(label, True, True, 5)

        image = Gtk.Image.new_from_icon_name('xsi-document-edit-symbolic', Gtk.IconSize.BUTTON)
        button = Gtk.Button(image=image, relief=Gtk.ReliefStyle.NONE, name='manager-group-edit-button')
        self.box.pack_end(button, False, False, 2)
        button.connect('clicked', self.edit_group_name)
        self.hoverbox.set_child_widget(button)

        self.box.show_all()

    def remove_group(self, *args):
        self.file_handler.remove_group(self.item.name)

    def edit_group_name(self, *args):
        self.box.destroy()
        self.box = None

        self.entry = Gtk.Entry(visible=True, text=self.item.name)
        self.hoverbox.add(self.entry)

        self.activate_id = self.entry.connect('activate', self.maybe_done)
        self.focus_id = self.entry.connect('focus-out-event', self.maybe_done)
        self.key_id = self.entry.connect('key-press-event', self.key_pressed)
        self.hoverbox.disable()

        self.entry.grab_focus()

    def maybe_done(self, *args):
        group_name = self.entry.get_text()
        if group_name != '' and group_name != self.item.name:
            old_name = self.item.name
            self.item.name = group_name
            self.file_handler.change_group_name(old_name, group_name)

        self.clean_up()

    def key_pressed(self, w, event):
        if event.keyval != Gdk.KEY_Escape:
            return Gdk.EVENT_PROPAGATE

        self.clean_up()

        return Gdk.EVENT_STOP

    def clean_up(self):
        if self.entry is None:
            return

        self.entry.disconnect(self.activate_id)
        self.entry.disconnect(self.focus_id)
        self.entry.disconnect(self.key_id)

        self.entry.destroy()
        self.entry = None

        self.generate_content()
        self.hoverbox.enable()

    def set_can_remove(self, can_remove):
        self.remove_item.set_sensitive(can_remove)

class Group(GObject.Object):
    def __init__(self, name, file_handler, model):
        super(Group, self).__init__()
        self.name = name
        self.file_handler = file_handler
        self.model = model

class Note(GObject.Object):
    def __init__(self, info, group_name):
        super(Note, self).__init__()
        self.info = info
        self.group_name = group_name
        self.text = info['text']
        if not 'title'in info or info['title'] in [None, '']:
            self.title = _("Untitled")
        else:
            self.title = info['title']

class NotesManager(object):
    def __init__(self, app, file_handler):
        self.app = app
        self.dragged_note = None
        self.search_model = Gio.ListStore()

        self.file_handler = file_handler
        self.file_handler.connect('group-changed', self.on_list_changed)
        self.file_handler.connect('lists-changed', self.generate_group_list)

        self.builder = Gtk.Builder()
        self.builder.set_translation_domain("sticky")
        self.builder.add_from_file('/usr/share/sticky/manager.ui')

        self.window = self.builder.get_object('main_window')
        self.group_list = self.builder.get_object('group_list')
        self.note_view = self.builder.get_object('note_view')
        self.note_view.connect('child-activated', self.on_note_activated)
        self.note_view.connect('selected-children-changed', self.on_selected_notes_changed)

        def create_group_entry(item):
            widget = GroupEntry(item)
            widget.drag_dest_set(Gtk.DestDefaults.MOTION | Gtk.DestDefaults.HIGHLIGHT, NOTE_TARGETS, Gdk.DragAction.MOVE)
            widget.connect('drag-drop', self.handle_drop)

            widget.new_item.connect('activate', self.new_group)
            return widget

        self.group_model = Gio.ListStore()
        self.group_list.bind_model(self.group_model, create_group_entry)

        search_toggle = self.builder.get_object('search_toggle_button')
        self.search_bar = self.builder.get_object('search_bar')
        GObject.Object.bind_property(search_toggle, 'active', self.search_bar, 'search_mode_enabled', GObject.BindingFlags.BIDIRECTIONAL)

        self.builder.get_object('new_note').connect('clicked', self.app.new_note)
        self.remove_note_button = self.builder.get_object('remove_note')
        self.remove_note_button.connect('clicked', self.remove_note)
        self.duplicate_note_button = self.builder.get_object('duplicate_note')
        self.duplicate_note_button.connect('clicked', self.duplicate_note)

        self.search_box = self.builder.get_object('search_box')
        self.search_box.connect('search-changed', self.on_search_changed)

        self.entry_box = self.builder.get_object('group_name_entry_box')
        self.entry_box.drag_dest_set(Gtk.DestDefaults.MOTION | Gtk.DestDefaults.HIGHLIGHT, NOTE_TARGETS, Gdk.DragAction.MOVE)
        self.entry_box.connect('drag-drop', self.handle_new_group_drop)

        main_menu = Gtk.Menu()

        item = Gtk.MenuItem(label=_("New Group"))
        item.connect('activate', self.new_group)
        main_menu.append(item)

        self.remove_group_item = Gtk.MenuItem(label=_("Remove Group"))
        self.remove_group_item.connect('activate', self.remove_group)
        main_menu.append(self.remove_group_item)

        main_menu.append(Gtk.SeparatorMenuItem(visible=True))

        item = Gtk.MenuItem(label=_("Back Up"))
        item.connect('activate', self.file_handler.save_backup)
        main_menu.append(item)

        item = Gtk.MenuItem(label=_("Restore Backups..."))
        item.connect('activate', self.file_handler.restore_backup, self.window)
        main_menu.append(item)

        main_menu.append(Gtk.SeparatorMenuItem(visible=True))

        item = Gtk.MenuItem(label=_("Import..."))
        item.connect('activate', self.file_handler.import_notes, self.window)
        main_menu.append(item)

        item = Gtk.MenuItem(label=_("Export..."))
        item.connect('activate', self.file_handler.export_notes, self.window)
        main_menu.append(item)

        main_menu.append(Gtk.SeparatorMenuItem(visible=True))

        item = Gtk.MenuItem(label=_("Preferences"))
        item.connect('activate', self.app.open_settings_window)
        main_menu.append(item)

        item = Gtk.MenuItem(label=_("Keyboard Shortcuts"))
        item.connect('activate', self.app.open_keyboard_shortcuts)
        main_menu.append(item)

        item = Gtk.MenuItem()
        item.set_label(_("About"))
        item.connect("activate", self.app.open_about)
        key, mod = Gtk.accelerator_parse("F1")
        accel_group = Gtk.AccelGroup()
        self.window.add_accel_group(accel_group)
        item.add_accelerator("activate", accel_group, key, mod, Gtk.AccelFlags.VISIBLE)
        main_menu.append(item)

        key, mod = Gtk.accelerator_parse('<Control>f')
        accel_group.connect(key, mod, Gtk.AccelFlags.VISIBLE, self.open_search)

        main_menu.show_all()

        self.builder.get_object('menu_button').set_popup(main_menu)

        self.generate_group_list()

        self.window.show_all()

        self.generate_previews()
        self.group_list.connect('row-selected', self.on_group_selected)
        self.group_list.connect('button-press-event', self.on_list_clicked)
        self.app.settings.connect('changed::active-group', self.on_active_group_changed)

    def on_list_clicked(self, list, event):
        for note in self.app.notes:
            note.present_with_time(Gtk.get_current_event_time())

    def on_list_changed(self, a, group_name):
        if group_name == self.get_current_group():
            self.generate_previews()

    def generate_group_list(self, *args):
        name = None
        selected_group_name = self.get_current_group()
        self.group_model.remove_all()

        for group_name in self.file_handler.get_note_group_names():
            model = Gio.ListStore()
            if self.app.settings.get_string('active-group') == group_name:
                name = group_name
            self.group_model.append(Group(group_name, self.file_handler, model))

        self.group_list.show_all()

        if name is not None:
            self.select_group(name)

        children = self.group_list.get_children()
        for item in children:
            item.set_can_remove(len(children) != 1)

        self.remove_group_item.set_sensitive(len(children) != 1)

    def select_group(self, name):
        for row in self.group_list.get_children():
            if row.item.name == name:
                self.group_list.select_row(row)
                break

    def on_active_group_changed(self, settings, key):
        value = settings.get_string(key)
        self.select_group(value)

    def on_group_selected(self, listbox, row):
        group_name = self.get_current_group()
        if group_name is not None:
            if self.app.settings.get_string('active-group') != group_name:
                self.app.settings.set_string('active-group', group_name)
            self.generate_previews()

    def on_search_changed(self, *args):
        search_text = self.search_box.get_text().lower().strip()
        if search_text == '':
            self.generate_previews()

        else:
            self.search_model.remove_all()

            for group_name in self.file_handler.get_note_group_names():
                for note_info in self.file_handler.get_note_list(group_name):
                    if note_info['title'].lower().find(search_text) != -1 or clean_text(note_info['text']).find(search_text) != -1:
                        self.search_model.append(Note(note_info, group_name))

            self.note_view.bind_model(self.search_model, self.create_note_entry)

    def open_search(self, *args):
        self.search_bar.set_search_mode(True)

    def on_note_activated(self, *args):
        activated = self.note_view.get_selected_children()[0].item
        activated_group = activated.group_name
        self.select_group(activated_group)

        self.app.focus_note(activated.info)

    def on_selected_notes_changed(self, *args):
        sensitive = len(self.note_view.get_selected_children()) != 0
        self.remove_note_button.set_sensitive(sensitive)
        self.duplicate_note_button.set_sensitive(sensitive)

    def create_note_entry(self, item):
        widget = Gtk.FlowBoxChild()
        widget.item = item

        dnd_wrapper = Gtk.EventBox(above_child=True)
        dnd_wrapper.drag_source_set(Gdk.ModifierType.BUTTON1_MASK, NOTE_TARGETS, Gdk.DragAction.MOVE)
        dnd_wrapper.connect('drag-begin', self.on_drag_begin)
        widget.add(dnd_wrapper)

        outer_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, margin=10, spacing=10)
        outer_box.set_receives_default(True)
        dnd_wrapper.add(outer_box)

        wrapper = Gtk.Box(halign=Gtk.Align.CENTER)
        context = wrapper.get_style_context()
        context.add_class(item.info['color'])
        context.add_class('note-preview')
        outer_box.pack_start(wrapper, False, False, 0)

        entry = NoteEntry(item, self.app.settings)
        wrapper.pack_start(entry, False, False, 0)

        widget.show_all()

        return widget

    def generate_previews(self, *args):
        selected_row = self.group_list.get_selected_row()
        if selected_row is None:
            return

        group_info = selected_row.item
        group_name = group_info.name
        model = group_info.model

        model.remove_all()

        for note in self.file_handler.get_note_list(group_name):
            model.append(Note(note, group_name))

        self.note_view.bind_model(model, self.create_note_entry)

    def get_current_group(self):
        row = self.group_list.get_selected_row()
        return row.item.name if row is not None else None

    def get_selected_note(self):
        return self.note_view.get_selected_children()[0].item.info

    def create_new_group(self, callback):
        entry = Gtk.Entry(visible=True)
        self.entry_box.pack_start(entry, False, False, 5)
        activate_id = 0
        focus_id = 0
        key_id = 0

        def clean_up(entry):
            entry.disconnect(activate_id)
            entry.disconnect(focus_id)
            entry.disconnect(key_id)

            self.entry_box.remove(entry)

        def maybe_done(entry, *args):
            group_name = entry.get_text()

            clean_up(entry)

            success = False
            if group_name != '':
                success = self.file_handler.new_group(group_name)

            if success:
                self.generate_group_list()

            callback(group_name, success)

        def key_pressed(entry, event):
            if event.keyval != Gdk.KEY_Escape:
                return Gdk.EVENT_PROPAGATE

            clean_up(entry)
            callback(None, False)

        entry.grab_focus()
        activate_id = entry.connect('activate', maybe_done)
        focus_id = entry.connect('focus-out-event', maybe_done)
        key_id = entry.connect('key-press-event', key_pressed)

    def new_group(self, *args):
        old_group = self.get_current_group()

        def on_complete(group_name, success):
            if not success:
                group_name = old_group

            for row in self.group_list.get_children():
                if row.item.name == group_name:
                    self.group_list.select_row(row)
                    return

        self.create_new_group(on_complete)

    def remove_note(self, *args):
        notes = []
        selected = self.get_selected_note()
        for child in self.note_view.get_children():
            if child.item.info != selected:
                notes.append(child.item.info)

        self.file_handler.update_note_list(notes, self.get_current_group())

    def remove_group(self, *args):
        group_name = self.get_current_group()

        self.file_handler.remove_group(group_name)

    def duplicate_note(self, *args):
        selected = self.get_selected_note()
        note_info = selected.copy()
        note_info['x'] += 50
        note_info['y'] += 50
        group = self.get_current_group()
        notes = self.file_handler.get_note_list(group)
        notes.append(note_info)
        self.file_handler.update_note_list(notes, group)

    def on_drag_begin(self, widget, *args):
        self.dragged_note = widget.get_parent().item.info

    def handle_drop(self, widget, context, x, y, time):
        new_group = widget.item.name
        new_list = self.file_handler.get_note_list(new_group)
        new_list.append(self.dragged_note)

        old_group = self.get_current_group()
        old_list = self.file_handler.get_note_list(old_group)
        old_list.remove(self.dragged_note)

        self.file_handler.update_note_list(new_list, new_group)
        self.file_handler.update_note_list(old_list, old_group)

        self.dragged_note = None

        Gtk.drag_finish(context, True, False, time)

    def handle_new_group_drop(self, widget, context, x, y, time):
        old_group = self.get_current_group()
        old_list = self.file_handler.get_note_list(old_group)

        def on_created(group_name, success):
            if not success:
                group_name = old_group
            else:
                old_list.remove(self.dragged_note)

                new_list = self.file_handler.get_note_list(group_name)
                new_list.append(self.dragged_note)

                self.file_handler.update_note_list(new_list, group_name)
                self.file_handler.update_note_list(old_list, old_group)

            self.dragged_note = None

            for row in self.group_list.get_children():
                if row.item.name == group_name:
                    self.group_list.select_row(row)
                    return

        self.create_new_group(on_created)

        Gtk.drag_finish(context, True, False, time)
